import numpy as np
import math
from typing import Tuple
from .algorithm import Algorithm

class OptimizedGradientMethod(Algorithm):
    def __init__(self, L, K):
        super().__init__(2, 1, [1], [1], [])
        self.L = L
        self.K = K
    
    def set_L(self, L):
        self.L = L
    
    def set_K(self, K):
        self.K = K
    
    def _compute_theta(self, k: int, K: int) -> float:
        if k < 0 or k > K:
            raise ValueError("k must be a non-negative integer and less than or equal to K.")
        
        theta = 1.0 
        for i in range(1, k + 1):
            if i == K:
                theta = (1 + math.sqrt(1 + 8 * theta ** 2)) / 2
            else:
                theta = (1 + math.sqrt(1 + 4 * theta ** 2)) / 2
        return theta
    
    def get_ABCD(self, k: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:

        if k < self.K:
            theta_k = self._compute_theta(k, self.K)
            theta_kp1 = self._compute_theta(k + 1, self.K)
            
            A = np.array([[1+(theta_k-1)/theta_kp1, (1-theta_k)/theta_kp1],
                          [1, 0]])
            
            B = np.array([[-(1+(2*theta_k-1)/theta_kp1)/self.L],
                          [-1/self.L]])
            
            C = np.array([[1, 0]])
            
            D = np.array([[0]])
        
        elif k == self.K:
            A = np.array([[0, 0],
                          [0, 0]])
            
            B = np.array([[0],
                          [0]])
            
            C = np.array([[1, 0]])
            
            D = np.array([[0]])
        
        else:
            raise ValueError("k must be less than or equal to K.")
        
        return (A, B, C, D)



import numpy as np
from typing import Tuple
from .algorithm import Algorithm

class ChambollePock(Algorithm):
    def __init__(self, tau, sigma, theta):
        super().__init__(2, 2, [1, 1], [1, 2], [])
        self.tau = tau
        self.sigma = sigma
        self.theta = theta
    
    def set_tau(self, tau):
        self.tau = tau
    
    def set_sigma(self, sigma):
        self.sigma = sigma
    
    def set_theta(self, theta):
        self.theta = theta
    
    def get_ABCD(self, k: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        A = np.array([[1, -self.tau],
                      [0, 0]])
        B = np.array([[-self.tau, 0],
                      [0, 1]])
        C = np.array([[1, -self.tau], 
                      [1, 1/self.sigma - self.tau*(1+self.theta)]])
        D = np.array([[-self.tau, 0], 
                      [-self.tau*(1+self.theta), -1/self.sigma]])
        return (A, B, C, D)



import numpy as np
from typing import Tuple
from .algorithm import Algorithm

class DouglasRachford(Algorithm):
    def __init__(self, gamma, lambda_value, operator_version = True):
        if operator_version:
            super().__init__(1, 2, [1, 1], [], [1, 2])
        else:
            super().__init__(1, 2, [1, 1], [1, 2], [])
        self.gamma = gamma
        self.lambda_value = lambda_value
    
    def set_gamma(self, gamma):
        self.gamma = gamma
    
    def set_lambda(self, lambda_value):
        self.lambda_value = lambda_value
    
    def get_ABCD(self, k: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        A = np.array([[1]])
        B = np.array([[-self.gamma*self.lambda_value, -self.gamma*self.lambda_value]])
        C = np.array([[1], 
                      [1]])
        D = np.array([[-self.gamma, 0], 
                      [-2*self.gamma, -self.gamma]])
        return (A, B, C, D)


import numpy as np
from typing import Tuple
from .algorithm import Algorithm

class Extragradient(Algorithm):
    def __init__(self, gamma: float, alpha: float, type: str = "unconstrained"):
        if type == "unconstrained":
            super().__init__(1, 1, [2], [], [1]) 
        else:
            raise ValueError("Not valid type/implemented yet")
        self.type = type
        self.gamma = gamma
        self.alpha = alpha
    
    def set_gamma(self, gamma):
        self.gamma = gamma
    
    def set_alpha(self, alpha):
        self.alpha = alpha
    
    def get_ABCD(self, k: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        if self.type == "unconstrained":
            A = np.array([[1]])
            B = np.array([[0, -self.alpha]])
            C = np.array([[1], [1]])
            D = np.array([[0, 0], 
                          [-self.gamma, 0]])
        else:
            raise ValueError("Not valid type/implemented yet")
        return (A, B, C, D)


import numpy as np
from typing import Tuple
from .algorithm import Algorithm

class TripleMomentum(Algorithm):
    # https://ieeexplore.ieee.org/document/7967721
    # Triple momentum method
    def __init__(self, sigma, beta):
        super().__init__(2, 1, [1], [1], [])
        self.sigma = sigma
        self.beta = beta
    
    def set_sigma(self, sigma):
        self.sigma = sigma

    def set_beta(self, beta):
        self.beta = beta

    def get_ABCD(self, k: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        q = self.sigma/self.beta
        
        alpha = (2-np.sqrt(q))/self.beta
        _beta = (1-np.sqrt(q))**2/(1+np.sqrt(q))
        gamma = (1-np.sqrt(q))**2/((2-np.sqrt(q))*(1+np.sqrt(q)))    
        
        A = np.array([[1+_beta,-_beta],[1,0]])
        B = np.array([[-alpha],[0]])
        C = np.array([[1+gamma,-gamma]])
        D = np.array([[0]])
        
        return (A, B, C, D)


import numpy as np
from typing import Tuple
from .algorithm import Algorithm

class TsengFBF(Algorithm):
    def __init__(self, gamma, theta):
        super().__init__(1, 2, [2, 1], [], [1, 2]) 
        self.gamma = gamma
        self.theta = theta
    
    def set_gamma(self, gamma):
        self.gamma = gamma
    
    def set_theta(self, theta):
        self.theta = theta
    
    def get_ABCD(self, k: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        A = np.array([[1]])
        B = np.array([[0, -self.gamma*self.theta, -self.gamma*self.theta]])
        C = np.array([[1],
                      [1],
                      [1]])
        D = np.array([[0, 0, 0], 
                      [-self.gamma, 0, -self.gamma],
                      [-self.gamma, 0, -self.gamma]])
        return (A, B, C, D)



import numpy as np
from mosek.fusion import Expr

def create_symmetric_matrix_expression(Xij, n):
    r"""
    Convert a list of upper triangle variables to a symmetric matrix expression.

    :param Xij: MOSEK variable containing the upper triangle and diagonal values.
    :type Xij: Variable
    :param n: Size of the symmetric matrix.
    :type n: int
    :return: Symmetric matrix expression of size n x n.
    :rtype: Expr
    """
    X_expr = [[None] * n for _ in range(n)]
    idx = 0
    for i in range(n):
        for j in range(i, n):
            X_expr[i][j] = Xij.index(idx)
            if i != j:
                X_expr[j][i] = Xij.index(idx)
            idx += 1
    X_rows = []
    for i in range(n):
        X_rows.append(Expr.hstack(X_expr[i]))
    X = Expr.vstack(X_rows)
    return X

def create_symmetric_matrix(upper_triangle_values, n):
    r"""
    Convert a list of upper triangle values to a symmetric matrix.

    :param upper_triangle_values: List of length n(n+1)/2 containing the upper triangle and diagonal values.
    :type upper_triangle_values: list
    :param n: Size of the symmetric matrix.
    :type n: int
    :return: Symmetric matrix of size n x n.
    :rtype: np.ndarray
    :raises ValueError: If the length of upper_triangle_values is not n(n+1)/2.
    """
    if len(upper_triangle_values) != n * (n + 1) // 2:
        raise ValueError("The length of upper_triangle_values must be n(n+1)/2")

    symmetric_matrix = np.zeros((n, n))

    idx = 0
    for i in range(n):
        for j in range(i, n):
            symmetric_matrix[i, j] = upper_triangle_values[idx]
            if i != j:
                symmetric_matrix[j, i] = upper_triangle_values[idx]
            idx += 1

    return symmetric_matrix



import numpy as np
from abc import ABC, abstractmethod
from typing import List, Tuple, Any, Dict, Union


class Algorithm(ABC):
    r"""
    Solves the problem:

    .. math::
       \text{find } y \in H \text{ such that } 0 \in \sum_{i \in I_{\text{func}}} \partial f_i(y) + \sum_{i \in I_{\text{op}}} G_i(y)

    with the state-space representation:

    .. math::
       x^{k+1} = (A_k \otimes I)x^k + (B_k \otimes I)u^k \quad \text{and} \quad
       y^k = (C_k \otimes I)x^k + (D_k \otimes I)u^k

    where:

    - **n** is the dimension of :math:`x`.
    - **m** is the total number of components (indices :math:`1,\dots, m`), split into
      :math:`I_{\text{func}}` and :math:`I_{\text{op}}`.
    - For each :math:`i`, **\bar{m}_i** is the number of evaluations and

      .. math::
         \bar{m} = \sum_{i=1}^{m} \bar{m}_i.
    """

    def __init__(self, n: int, m: int, m_bar_is: List[int],
                 I_func: List[int], I_op: List[int]):
        # Basic validations
        if n < 1:
            raise ValueError("n must be at least 1")
        if m < 1:
            raise ValueError("m must be at least 1")
        if m != len(m_bar_is):
            raise ValueError("m must equal the length of m_bar_is")
        if not set(I_func).isdisjoint(I_op):
            raise ValueError("I_func and I_op must be disjoint")
        if set(I_func).union(I_op) != set(range(1, m + 1)):
            raise ValueError("I_func and I_op must cover {1,…, m}")
        if I_func and any(I_func[i] >= I_func[i + 1] for i in range(len(I_func) - 1)):
            raise ValueError("I_func must be in increasing order")
        if I_op and any(I_op[i] >= I_op[i + 1] for i in range(len(I_op) - 1)):
            raise ValueError("I_op must be in increasing order")

        self.n = n                      # Dimension of x.
        self.m = m                      # Total components.
        self.m_bar_is = m_bar_is        # Evaluations per component (\bar{m}_i).
        self.m_bar = sum(m_bar_is)      # Total evaluations (\bar{m}).
        self.I_func = I_func            # Functional indices (I_{\text{func}}).
        self.I_op = I_op                # Operator indices (I_{\text{op}}).
        self.m_func = len(I_func)       # Count of functional components.
        self.m_op = len(I_op)           # Count of operator components.
        self.m_bar_func = sum(m_bar_is[i - 1] for i in I_func) if I_func else 0  # \bar{m}_{\text{func}}
        self.m_bar_op = sum(m_bar_is[i - 1] for i in I_op) if I_op else 0          # \bar{m}_{\text{op}}
        # Mapping for functional indices (for F matrices)
        self.kappa = {I_func[i]: i + 1 for i in range(self.m_func)} if I_func else {}

    @abstractmethod
    def get_ABCD(self, k: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        r"""
        Return the system matrices (A, B, C, D) for iteration k.

        **Dimensions:**

          - **A**: (n x n)
          - **B**: (n x \bar{m})
          - **C**: (\bar{m} x n)
          - **D**: (\bar{m} x \bar{m})

        :param k: Iteration index.
        :type k: int
        :returns: A tuple (A, B, C, D) of numpy arrays.
        :rtype: Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
        """
        pass

    def get_AsBsCsDs(self, k_min: int, k_max: int
                      ) -> Dict[int, Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]]:
        r"""
        Return a dictionary mapping each iteration index k to the tuple (A, B, C, D)
        for k_min ≤ k ≤ k_max.

        :param k_min: The minimum iteration index.
        :type k_min: int
        :param k_max: The maximum iteration index.
        :type k_max: int
        :returns: A dictionary where keys are iteration indices and values are tuples (A, B, C, D).
        :rtype: Dict[int, Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]]
        """
        if k_min < 0 or k_max < k_min:
            raise ValueError("Require 0 <= k_min <= k_max")
        return {k: self.get_ABCD(k) for k in range(k_min, k_max + 1)}

    # --- U MATRICES ---
    def _generate_U(self, k_min: int, k_max: int, k: int = None, star: bool = False) -> np.ndarray:
        r"""
        Generate a U matrix.

        The total number of columns is given by:

        .. math::
           n + ((k_{max} - k_{min} + 1) \cdot \bar{m} + m).

        - If ``star=True``, returns U_{\text{star}} defined as:

          .. math::
             U_{\text{star}} = \begin{bmatrix}
             \mathbf{0}_{m \times \left(n + ((k_{max} - k_{min} + 1) \cdot \bar{m})\right)} &
             N &
             \mathbf{0}_{m \times 1}
             \end{bmatrix},

          where

          .. math::
             N = \begin{bmatrix} I_{m-1} \\ -\mathbf{1}_{1 \times (m-1)} \end{bmatrix}.

        - If ``star=False`` and k is provided (with k_min ≤ k ≤ k_max), returns:

          .. math::
             U_k = \begin{bmatrix}
             \mathbf{0}_{\bar{m} \times \left(n + (k - k_{min})\bar{m}\right)} &
             I_{\bar{m}} &
             \mathbf{0}_{\bar{m} \times \left((k_{max} - k)\bar{m} + m\right)}
             \end{bmatrix}.

        :param k_min: The minimum iteration index.
        :type k_min: int
        :param k_max: The maximum iteration index.
        :type k_max: int
        :param k: (Optional) The current iteration index (required if star is False).
        :type k: int, optional
        :param star: If True, generate the star matrix.
        :type star: bool
        :returns: The generated U matrix.
        :rtype: np.ndarray
        """
        total_cols = self.n + ((k_max - k_min + 1) * self.m_bar + self.m)
        if star:
            if self.m > 1:
                N = np.vstack([np.eye(self.m - 1), -np.ones((1, self.m - 1))])
                return np.hstack([
                    np.zeros((self.m, self.n + (k_max - k_min + 1) * self.m_bar)),
                    N,
                    np.zeros((self.m, 1))
                ])
            else:
                return np.zeros((self.m, total_cols))
        else:
            if k is None:
                raise ValueError("When star is False, k must be provided")
            if k < k_min or k > k_max:
                raise ValueError("k must be between k_min and k_max")
            left = np.zeros((self.m_bar, self.n + (k - k_min) * self.m_bar))
            ident = np.eye(self.m_bar)
            right = np.zeros((self.m_bar, (k_max - k) * self.m_bar + self.m))
            return np.hstack([left, ident, right])

    def get_Us(self, k_min: int, k_max: int) -> Dict[Any, np.ndarray]:
        r"""
        Return a dictionary of U matrices for iterations k_min ≤ k ≤ k_max, including the star matrix.

        :param k_min: The minimum iteration index.
        :type k_min: int
        :param k_max: The maximum iteration index.
        :type k_max: int
        :returns: A dictionary where keys are iteration indices (and 'star') and values are U matrices.
        :rtype: Dict[Any, np.ndarray]
        """
        if k_min < 0 or k_max < k_min:
            raise ValueError("Require 0 <= k_min <= k_max")
        Us = {}
        for k in range(k_min, k_max + 1):
            Us[k] = self._generate_U(k_min, k_max, k=k, star=False)
        Us['star'] = self._generate_U(k_min, k_max, star=True)
        return Us

    # --- Y MATRICES ---
    def _generate_Y(self,
                    sys_mats: Dict[int, Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]],
                    k_min: int, k_max: int, k: int = None, star: bool = False) -> np.ndarray:
        r"""
        Generate the output matrix Y using system matrices from sys_mats.

        The total number of columns is:

        .. math::
           n + (k_{max} - k_{min} + 1) \cdot \bar{m} + m.

        - If ``star=True``, returns Y_{\text{star}} defined as:

          .. math::
             Y_{\text{star}} = \begin{bmatrix} \mathbf{0}_{m \times (total\_cols - 1)} & \mathbf{1}_{m \times 1} \end{bmatrix}.

        - Otherwise, if star is False then k must be provided:
          
          - If :math:`k = k_{min}`, then

            .. math::
               Y_{k_{min}} = \begin{bmatrix} C_{k_{min}} & D_{k_{min}} & \mathbf{0} \end{bmatrix},

            where the zeros block has shape :math:`(\bar{m}, ((k_{max} - k_{min}) \cdot \bar{m} + m))`.

          - If :math:`k = k_{min} + 1`, then

            .. math::
               Y_{k_{min}+1} = \begin{bmatrix} C_{k_{min}+1} A_{k_{min}} & C_{k_{min}+1} B_{k_{min}} & D_{k_{min}+1} & \mathbf{0} \end{bmatrix},

            with the zeros block of shape :math:`(\bar{m}, ((k_{max} - k_{min} - 1) \cdot \bar{m} + m))`.

          - If :math:`k \ge k_{min} + 2`, then construct Y_k by:
            - Block 1: :math:`C_k (A_{k-1} \cdots A_{k_{min}})`,
            - For each :math:`j` from :math:`k_{min}` to :math:`k-2`, the next block is
              :math:`C_k (A_{k-1} \cdots A_{j+1}) B_j`,
            - Then block: :math:`C_k B_{k-1}`,
            - Followed by :math:`D_k` and a zeros block of shape :math:`(\bar{m}, ((k_{max} - k) \cdot \bar{m} + m))`.

        :param sys_mats: Dictionary mapping each iteration index to (A, B, C, D).
        :type sys_mats: Dict[int, Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]]
        :param k_min: The minimum iteration index.
        :type k_min: int
        :param k_max: The maximum iteration index.
        :type k_max: int
        :param k: (Optional) The current iteration index (required if star is False).
        :type k: int, optional
        :param star: If True, generate Y_{\text{star}}.
        :type star: bool
        :returns: The generated Y matrix.
        :rtype: np.ndarray
        """
        total_cols = self.n + (k_max - k_min + 1) * self.m_bar + self.m

        if star:
            return np.hstack([
                np.zeros((self.m, total_cols - 1)),
                np.ones((self.m, 1))
            ])
        if k is None:
            raise ValueError("When star is False, k must be provided")

        if k == k_min:
            C_kmin = sys_mats[k_min][2]
            D_kmin = sys_mats[k_min][3]
            zeros_blk = np.zeros((self.m_bar, (k_max - k_min) * self.m_bar + self.m))
            return np.hstack([C_kmin, D_kmin, zeros_blk])
        elif k == k_min + 1:
            C_next = sys_mats[k][2]
            A_kmin = sys_mats[k_min][0]
            B_kmin = sys_mats[k_min][1]
            D_next = sys_mats[k][3]
            zeros_blk = np.zeros((self.m_bar, (k_max - k_min - 1) * self.m_bar + self.m))
            block1 = C_next @ A_kmin
            block2 = C_next @ B_kmin
            return np.hstack([block1, block2, D_next, zeros_blk])
        else:
            blocks = []
            C_k = sys_mats[k][2]
            prod = C_k.copy()
            for i in reversed(range(k_min, k)):
                prod = prod @ sys_mats[i][0]
            blocks.append(prod)
            for j in range(k_min, k - 1):
                prod_B = np.eye(self.n)
                for i in reversed(range(j + 1, k)):
                    prod_B = prod_B @ sys_mats[i][0]
                blocks.append(C_k @ prod_B @ sys_mats[j][1])
            blocks.append(C_k @ sys_mats[k - 1][1])
            D_k = sys_mats[k][3]
            blocks.append(D_k)
            zeros_blk = np.zeros((self.m_bar, (k_max - k) * self.m_bar + self.m))
            blocks.append(zeros_blk)
            return np.hstack(blocks)

    def get_Ys(self, k_min: int, k_max: int) -> Dict[Any, np.ndarray]:
        r"""
        Return a dictionary of Y matrices for iterations k_min ≤ k ≤ k_max, including Y_{\text{star}}.

        :param k_min: The minimum iteration index.
        :type k_min: int
        :param k_max: The maximum iteration index.
        :type k_max: int
        :returns: A dictionary where keys are iteration indices (and 'star') and values are Y matrices.
        :rtype: Dict[Any, np.ndarray]
        """
        if k_min < 0 or k_max < k_min:
            raise ValueError("Require 0 <= k_min <= k_max")
        sys_mats = self.get_AsBsCsDs(k_min, k_max)
        Ys = {}
        for k in range(k_min, k_max + 1):
            Ys[k] = self._generate_Y(sys_mats, k_min, k_max, k=k, star=False)
        Ys['star'] = self._generate_Y(sys_mats, k_min, k_max, star=True)
        return Ys

    # --- X MATRICES ---
    def _generate_X_k(self, sys_mats: Dict[int, Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]],
                        k: int, k_min: int, k_max: int) -> np.ndarray:
        r"""
        Generate the state matrix X_k for iterations k_min ≤ k ≤ k_max+1.

        - For :math:`k = k_{min}`, X_{k_{min}} is given by
          .. math:: [I_n, \mathbf{0}],
          where the zeros block has the appropriate number of columns.
        - For :math:`k = k_{min} + 1`, X_{k_{min}+1} is given by
          .. math:: [A_{k_{min}}, B_{k_{min}}, \mathbf{0}],
          with a zeros block of appropriate size.
        - For :math:`k \ge k_{min} + 2`, X_k is constructed as:

          .. math::
             X_k = \left[ A_{k-1}\cdots A_{k_{min}},\; (A_{k-1}\cdots A_{j+1}) B_j \text{ for } j=k_{min},\dots,k-2,\; B_{k-1},\; \mathbf{0} \right].

        :param sys_mats: Dictionary mapping each iteration index to (A, B, C, D).
        :type sys_mats: Dict[int, Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]]
        :param k: The current iteration index.
        :type k: int
        :param k_min: The minimum iteration index.
        :type k_min: int
        :param k_max: The maximum iteration index.
        :type k_max: int
        :returns: The generated X_k matrix.
        :rtype: np.ndarray
        """
        total_cols = self.n + (k_max - k_min + 1) * self.m_bar + self.m

        if k < k_min or k > k_max + 1:
            raise ValueError("k must be in [k_min, k_max+1]")

        if k == k_min:
            return np.hstack([np.eye(self.n), np.zeros((self.n, total_cols - self.n))])
        if k == k_min + 1:
            A, B, _, _ = sys_mats[k_min]
            return np.hstack([A, B, np.zeros((self.n, total_cols - self.n - self.m_bar))])
        
        parts = []
        prod = np.eye(self.n)
        for i in reversed(range(k_min, k)):
            prod = prod @ sys_mats[i][0]
        parts.append(prod)
        for j in range(k_min, k - 1):
            prod_B = np.eye(self.n)
            for i in reversed(range(j + 1, k)):
                prod_B = prod_B @ sys_mats[i][0]
            parts.append(prod_B @ sys_mats[j][1])
        parts.append(sys_mats[k - 1][1])
        zeros_blk = np.zeros((self.n, (k_max + 1 - k) * self.m_bar + self.m))
        parts.append(zeros_blk)
        return np.hstack(parts)

    def get_Xs(self, k_min: int, k_max: int) -> Dict[int, np.ndarray]:
        r"""
        Return a dictionary mapping each iteration index k (for k_min ≤ k ≤ k_max+1) to the corresponding X_k matrix.

        The X_k matrices are constructed using system matrices retrieved via
        :meth:`get_AsBsCsDs` and the helper method :meth:`_generate_X_k`.

        :param k_min: The minimum iteration index.
        :type k_min: int
        :param k_max: The maximum iteration index.
        :type k_max: int
        :returns: A dictionary with keys as iteration indices and values as X_k matrices.
        :rtype: Dict[int, np.ndarray]
        """
        if k_min < 0 or k_max < k_min:
            raise ValueError("Require 0 <= k_min <= k_max")
        sys_mats = self.get_AsBsCsDs(k_min, k_max)
        Xs = {}
        for k in range(k_min, k_max + 2):
            Xs[k] = self._generate_X_k(sys_mats, k, k_min, k_max)
        return Xs

    # --- PROJECTION MATRICES (P) ---
    def get_Ps(self) -> Dict[Tuple[Any, Any], np.ndarray]:
        r"""
        Return a dictionary of projection matrices P.

        For each component index :math:`i \in \{1,\dots, m\}` and for each evaluation index
        :math:`j \in \{1,\dots, \bar{m}_i\}`, the projection matrix :math:`P_{(i,j)}` is a
        1 x \bar{m} row vector with a 1 at the (offset+j)-th position, where

        .. math::
           \text{offset} = \sum_{l=1}^{i-1} \bar{m}_l.

        Additionally, :math:`P_{(i,\text{'star'})}` is a 1 x m row vector with a 1 in the i-th position.

        :returns: A dictionary where keys are tuples (i, j) or (i, 'star') and values are projection matrices.
        :rtype: Dict[Tuple[Any, Any], np.ndarray]
        """
        Ps = {}
        for i in range(1, self.m + 1):
            offset = sum(self.m_bar_is[:i - 1])
            for j in range(1, self.m_bar_is[i - 1] + 1):
                vec = np.zeros((1, self.m_bar))
                vec[0, offset + j - 1] = 1
                Ps[(i, j)] = vec
            star_vec = np.zeros((1, self.m))
            star_vec[0, i - 1] = 1
            Ps[(i, 'star')] = star_vec
        return Ps

    # --- F MATRICES (for functional components) ---
    def _generate_F(self, i: int, j: int = None, k: int = None,
                    star: bool = False, k_min: int = 0, k_max: int = 0) -> np.ndarray:
        r"""
        Generate a single row of the F matrix for a functional component indexed by i.

        The overall F row has dimension

        .. math::
           \left( 1, \, ((k_{max} - k_{min} + 1) \cdot \bar{m}_{\text{func}} + m_{\text{func}}) \right).

        - For the non-star case (i.e. :math:`F_{(i,j,k)}`), a 1 is placed at the location corresponding to
          the j-th evaluation of component i, shifted by the contributions of all preceding functional components
          and by (k - k_{min}) blocks of size \(\bar{m}_{\text{func}}\).
        - For the star case (i.e. :math:`F_{(i,\text{'star'},\text{'star'})}`), a 1 is placed in the last
          m_{\text{func}} entries, specifically at index

          .. math::
             (k_{max} - k_{min} + 1) \cdot \bar{m}_{\text{func}} + (\kappa[i]-1).

        :param i: Functional component index (must be in :attr:`I_func`).
        :type i: int
        :param j: Evaluation index for the non-star case.
        :type j: int, optional
        :param k: Iteration index for the non-star case (must be in [k_min, k_max]).
        :type k: int, optional
        :param star: If True, generate the star row.
        :type star: bool
        :param k_min: The minimum iteration index.
        :type k_min: int
        :param k_max: The maximum iteration index.
        :type k_max: int
        :returns: A 1 x total_dim numpy array representing the generated F row.
        :rtype: np.ndarray
        """
        total_dim = ((k_max - k_min + 1) * self.m_bar_func + self.m_func)

        if not self.I_func or i not in self.I_func:
            raise ValueError(f"i must be in I_func. Got i = {i}")
        idx = self.kappa[i]  # 1-indexed

        if star:
            if j is not None or k is not None:
                raise ValueError("For star F matrices, do not supply j or k")
            F_star = np.zeros((1, total_dim))
            F_star[0, (k_max - k_min + 1) * self.m_bar_func + idx - 1] = 1
            return F_star
        else:
            if j is None or k is None:
                raise ValueError("For non-star F matrices, both j and k must be provided")
            if not (k_min <= k <= k_max):
                raise ValueError(f"k must be in [{k_min}, {k_max}]. Got k = {k}")
            max_j = self.m_bar_is[i - 1]
            if not (1 <= j <= max_j):
                raise ValueError(f"For i = {i}, j must be in [1, {max_j}]. Got j = {j}")
            F_nonstar = np.zeros((1, total_dim))
            offset = sum(self.m_bar_is[l - 1] for l in self.I_func if l < i)
            start_idx = self.m_bar_func * (k - k_min) + offset
            F_nonstar[0, start_idx + j - 1] = 1
            return F_nonstar

    def get_Fs(self, k_min: int, k_max: int) -> Dict[Tuple[Any, Any, Any], np.ndarray]:
        r"""
        Return a dictionary of F matrices for all functional components.

        The dictionary keys are defined as follows:
        - For non-star F matrices: keys are of the form (i, j, k), where
          - i is in :attr:`I_func`,
          - j is in [1, m_bar_is[i-1]],
          - k is in [k_min, k_max].
        - For star F matrices: keys are of the form (i, 'star', 'star') for each i in :attr:`I_func`.

        :param k_min: The minimum iteration index.
        :type k_min: int
        :param k_max: The maximum iteration index.
        :type k_max: int
        :returns: A dictionary mapping keys to the corresponding F row matrices.
        :rtype: Dict[Tuple[Any, Any, Any], np.ndarray]
        """
        if k_min < 0 or k_max < k_min:
            raise ValueError("Require 0 <= k_min <= k_max")
        Fs = {}
        for k in range(k_min, k_max + 1):
            for i in self.I_func:
                for j in range(1, self.m_bar_is[i - 1] + 1):
                    Fs[(i, j, k)] = self._generate_F(i, j, k, star=False,
                                                      k_min=k_min, k_max=k_max)
        for i in self.I_func:
            Fs[(i, 'star', 'star')] = self._generate_F(i, star=True,
                                                         k_min=k_min, k_max=k_max)
        return Fs

    # --- LIFTED CONSTRAINT MATRICES ---
    def compute_E(self, i: int, pairs: List[Tuple[Any, Any]],
                  k_min: int, k_max: int) -> np.ndarray:
        r"""
        Compute the E matrix for component i using a list of (j, k) pairs.

        The E matrix is defined as:

        .. math::
           E^{k_{min}, k_{max}}_{(i, j_1, k_1, \dots, j_{n_{i,o}}, k_{n_{i,o}})} =
           \begin{bmatrix}
           P_{(i,j_1)}Y_{k_1} \\
           \vdots \\
           P_{(i,j_{n_{i,o}})}Y_{k_{n_{i,o}}} \\
           P_{(i,j_1)}U_{k_1} \\
           \vdots \\
           P_{(i,j_{n_{i,o}})}U_{k_{n_{i,o}}}
           \end{bmatrix}.

        The resulting matrix has dimensions

        .. math::
           2 \cdot (\text{number of pairs}) \times \left[ n + (k_{max} - k_{min} + 1) \cdot \bar{m} + m \right].

        :param i: The component index.
        :type i: int
        :param pairs: A list of (j, k) pairs. For non-star pairs, j must be an integer in
                      [1, m_bar_is[i-1]] and k must lie in [k_min, k_max]. For the star case,
                      the pair should be ('star', 'star').
        :type pairs: List[Tuple[Any, Any]]
        :param k_min: The minimum iteration index.
        :type k_min: int
        :param k_max: The maximum iteration index.
        :type k_max: int
        :returns: The computed E matrix.
        :rtype: np.ndarray
        """
        for pair in pairs:
            if not (isinstance(pair, tuple) and len(pair) == 2):
                raise ValueError("Each element in pairs must be a tuple of two elements (j, k).")
        if k_min < 0 or k_max < k_min:
            raise ValueError("Invalid iteration bounds: ensure 0 <= k_min <= k_max.")

        Ps = self.get_Ps()
        Ys = self.get_Ys(k_min, k_max)
        Us = self.get_Us(k_min, k_max)

        Y_blocks = []
        U_blocks = []

        for (j, k) in pairs:
            if j == 'star' and k == 'star':
                P = Ps[(i, 'star')]
                Y_blk = P @ Ys['star']
                U_blk = P @ Us['star']
            else:
                if not (isinstance(j, int) and isinstance(k, int)):
                    raise ValueError("For non-star pairs, (j, k) must be integers.")
                if not (1 <= j <= self.m_bar_is[i - 1]):
                    raise ValueError(f"For i = {i}, j must be in [1, {self.m_bar_is[i - 1]}].")
                if not (k_min <= k <= k_max):
                    raise ValueError(f"k must be in [{k_min}, {k_max}].")
                P = Ps[(i, j)]
                Y_blk = P @ Ys[k]
                U_blk = P @ Us[k]
            Y_blocks.append(Y_blk)
            U_blocks.append(U_blk)

        return np.vstack(Y_blocks + U_blocks)

    def compute_W(self, i: int, pairs: List[Tuple[Any, Any]],
                  k_min: int, k_max: int, M: np.ndarray) -> np.ndarray:
        r"""
        Compute the W matrix for component i.

        The W matrix is given by:

        .. math::
           W = E^T M E,

        where E is computed via :meth:`compute_E`.

        :param i: The component index.
        :type i: int
        :param pairs: A list of (j, k) pairs. For non-star pairs, each pair must be (int, int)
                      with j in the valid range for component i and k in [k_min, k_max]. For star rows,
                      the pair should be ('star', 'star').
        :type pairs: List[Tuple[Any, Any]]
        :param k_min: The minimum iteration index.
        :type k_min: int
        :param k_max: The maximum iteration index.
        :type k_max: int
        :param M: A symmetric matrix of dimension [2*(number of pairs) x 2*(number of pairs)].
        :type M: np.ndarray
        :returns: The computed W matrix.
        :rtype: np.ndarray
        :raises ValueError: If any input conditions are not met.
        """
        for pair in pairs:
            if not (isinstance(pair, tuple) and len(pair) == 2):
                raise ValueError("Each element in pairs must be a tuple of two elements (j, k).")
        if k_min < 0 or k_max < k_min:
            raise ValueError("Invalid iteration bounds: ensure 0 <= k_min <= k_max.")
        exp_dim = 2 * len(pairs)
        if M.ndim != 2 or M.shape != (exp_dim, exp_dim):
            raise ValueError(f"M must be a square matrix of dimension [{exp_dim} x {exp_dim}].")
        if not np.allclose(M, M.T, atol=1e-8):
            raise ValueError("M must be symmetric.")

        E = self.compute_E(i, pairs, k_min, k_max)
        return E.T @ M @ E

    def compute_F_aggregated(self, i: int, pairs: List[Union[Tuple[int, int], Tuple[str, str]]],
                             k_min: int, k_max: int, a: np.ndarray) -> np.ndarray:
        r"""
        Compute the aggregated F vector for component i.

        The aggregated F vector is computed as:

        .. math::
           F_{\text{aggregated}} = \begin{bmatrix} (F_{(i,j_1,k_1)})^T & \cdots & (F_{(i,j_n,k_n)})^T \end{bmatrix} a,

        where each F row is obtained from the F matrices (via :meth:`get_Fs`), transposed, and
        horizontally stacked. The resulting matrix has shape (total_dim, number of pairs) and is then
        multiplied by the weight vector :math:`a` (a 1D array of length equal to the number of pairs) to
        yield a column vector of shape (total_dim, 1).

        Here,

        .. math::
           \text{total\_dim} = n + (k_{max} - k_{min} + 1) \cdot \bar{m}_{\text{func}} + m_{\text{func}}.

        :param i: The component index (should be in :attr:`I_func`).
        :type i: int
        :param pairs: A list of (j, k) pairs. For non-star cases, each pair must be (int, int) with
                      j in [1, m_bar_is[i-1]] and k in [k_min, k_max]. For the star case, the pair
                      should be ('star', 'star').
        :type pairs: List[Union[Tuple[int, int], Tuple[str, str]]]
        :param k_min: The minimum iteration index.
        :type k_min: int
        :param k_max: The maximum iteration index.
        :type k_max: int
        :param a: A 1D numpy array of weights with length equal to the number of pairs.
        :type a: np.ndarray
        :returns: The aggregated F vector as a column vector.
        :rtype: np.ndarray
        :raises ValueError: If any input conditions are not met.
        """
        if i not in self.I_func:
            raise ValueError(f"Component index i must be in I_func. Got i = {i}.")

        total_dim = (k_max - k_min + 1) * self.m_bar_func + self.m_func

        for pair in pairs:
            if not (isinstance(pair, tuple) and len(pair) == 2):
                raise ValueError("Each element in pairs must be a tuple of two elements (j, k).")
            j, k = pair
            if (j == 'star' or k == 'star') and not (j == 'star' and k == 'star'):
                raise ValueError("For star F matrices, both j and k must be 'star'.")
            if j != 'star' and k != 'star':
                if not (isinstance(j, int) and isinstance(k, int)):
                    raise ValueError("For non-star F matrices, both j and k must be integers.")
                if not (k_min <= k <= k_max):
                    raise ValueError(f"k must be in [{k_min}, {k_max}]. Got k = {k}.")
                if not (1 <= j <= self.m_bar_is[i - 1]):
                    raise ValueError(f"For component i = {i}, j must be in [1, {self.m_bar_is[i - 1]}]. Got j = {j}.")
        if not pairs:
            raise ValueError("Pairs list must be nonempty.")

        if not (isinstance(a, np.ndarray) and a.ndim == 1 and len(a) == len(pairs)):
            raise ValueError("a must be a 1D numpy array with length equal to the number of pairs.")

        Fs_dict = self.get_Fs(k_min, k_max)
        F_cols = []
        for (j, k) in pairs:
            key = (i, 'star', 'star') if (j == 'star' and k == 'star') else (i, j, k)
            if key not in Fs_dict:
                raise ValueError(f"Key {key} not found in F matrices.")
            F_cols.append(Fs_dict[key].T)
        F_stack = np.hstack(F_cols)
        aggregated_F = F_stack @ a
        return aggregated_F.reshape(total_dim, 1)



import numpy as np
from typing import Tuple
from .algorithm import Algorithm

class GradientNesterovMomentum(Algorithm):
    def __init__(self, gamma, delta):
        super().__init__(2, 1, [1], [1], [])
        self.gamma = gamma
        self.delta = delta
    
    def set_gamma(self, gamma):
        self.gamma = gamma

    def set_delta(self, delta):
        self.delta = delta
    
    def get_ABCD(self, k: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        A = np.array([[1+self.delta,-self.delta],[1,0]])
        B = np.array([[-self.gamma],[0]])
        C = np.array([[1+self.delta,-self.delta]])
        D = np.array([[0]])
        return (A, B, C, D)



import numpy as np
from typing import Tuple
from .algorithm import Algorithm

class GradientMethod(Algorithm):
    def __init__(self, gamma):
        super().__init__(n=1, m=1, m_bar_is=[1], I_func=[1], I_op=[])
        self.gamma = gamma
    
    def set_gamma(self, gamma):
        self.gamma = gamma
    
    def get_ABCD(self, k: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        A = np.array([[1]])
        B = np.array([[-self.gamma]])
        C = np.array([[1]])
        D = np.array([[0]])
        return (A, B, C, D)


import numpy as np
from abc import ABC, abstractmethod
from typing import List, Union, Tuple, Dict

# ---------------------------------------------------------------------------
# InterpolationIndices Class
# ---------------------------------------------------------------------------
class InterpolationIndices:
    r"""
    A wrapper for interpolation indices that ensures only allowed values are used.

    :param value: The interpolation index string.
    :type value: str
    :raises ValueError: If the provided value is not in the allowed set.
    """
    ALLOWED_VALUES = {"i<j", "i!=j", "i", "i!=star"}

    def __init__(self, value: str):
        if value not in self.ALLOWED_VALUES:
            raise ValueError(f"Invalid interpolation index: {value}. Allowed values: {self.ALLOWED_VALUES}")
        self.value = value

    def __str__(self):
        return self.value

    def __eq__(self, other):
        if isinstance(other, InterpolationIndices):
            return self.value == other.value
        elif isinstance(other, str):
            return self.value == other
        return False

    def __repr__(self):
        return f"InterpolationIndices({self.value})"


# ---------------------------------------------------------------------------
# Abstract Base Classes for Interpolation Conditions
# ---------------------------------------------------------------------------
class InterpolationCondition(ABC):
    r"""
    Abstract base class for an interpolation condition.

    Derived classes must implement the :meth:`get_data` method, which returns a list of tuples 
    representing the interpolation data.
    """
    @abstractmethod
    def get_data(self):
        r"""
        Return the interpolation data.

        :returns: A list of tuples representing the interpolation data.
        :rtype: list
        """
        pass

class OperatorInterpolationCondition(InterpolationCondition):
    r"""
    Base class for operator interpolation conditions.

    Must return a list of tuples of the form:
    
    ``(matrix, interpolation_indices)``

    where:

    - **matrix** is a square, symmetric 2D numpy array.
    - **interpolation_indices** is an instance of :class:`InterpolationIndices`.
    """
    @abstractmethod
    def get_data(self) -> List[Tuple[np.ndarray, InterpolationIndices]]:
        r"""
        Return the operator interpolation data.

        :returns: A list of tuples, each containing a square symmetric matrix and an instance of :class:`InterpolationIndices`.
        :rtype: List[Tuple[np.ndarray, InterpolationIndices]]
        """
        pass

class FunctionInterpolationCondition(InterpolationCondition):
    r"""
    Base class for function interpolation conditions.

    Must return a list of tuples of the form:
    
    ``(matrix, vector, eq, interpolation_indices)``

    where:

    - **matrix** is a square, symmetric 2D numpy array.
    - **vector** is a 1D numpy array.
    - **eq** is a boolean flag (True for equality, False for inequality).
    - **interpolation_indices** is an instance of :class:`InterpolationIndices`.
    """
    @abstractmethod
    def get_data(self) -> List[Tuple[np.ndarray, np.ndarray, bool, InterpolationIndices]]:
        r"""
        Return the function interpolation data.

        :returns: A list of tuples, each containing a square symmetric matrix, a 1D vector, a boolean flag,
                  and an instance of :class:`InterpolationIndices`.
        :rtype: List[Tuple[np.ndarray, np.ndarray, bool, InterpolationIndices]]
        """
        pass


# ---------------------------------------------------------------------------
# Concrete Operator Interpolation Condition Classes
# ---------------------------------------------------------------------------
class MaximallyMonotone(OperatorInterpolationCondition):
    r"""
    Operator interpolation condition for a maximally monotone operator.

    This condition does not require any parameters.
    """
    def get_data(self) -> List[Tuple[np.ndarray, InterpolationIndices]]:
        r"""
        Return the interpolation data for a maximally monotone operator.

        :returns: A list containing one tuple with the interpolation matrix and indices.
        :rtype: List[Tuple[np.ndarray, InterpolationIndices]]
        """
        matrix = 0.5 * np.array([
            [0, 0, -1, 1],
            [0, 0,  1, -1],
            [-1, 1,  0, 0],
            [1, -1,  0, 0]
        ])
        interp_idx = InterpolationIndices("i<j")
        return [(matrix, interp_idx)]

class StronglyMonotone(OperatorInterpolationCondition):
    r"""
    Operator interpolation condition for a strongly monotone operator.

    :param mu: The strong monotonicity parameter (must be > 0 and finite).
    :type mu: float
    :raises ValueError: If mu is not a number, <= 0, or infinite.
    """
    def __init__(self, mu: Union[int, float]):
        if not isinstance(mu, (int, float)):
            raise ValueError("Strong monotonicity parameter must be a number.")
        mu = float(mu)
        if mu <= 0:
            raise ValueError("Strong monotonicity parameter (mu) must be greater than 0.")
        if mu == float('inf'):
            raise ValueError("Strong monotonicity parameter (mu) must be finite.")
        self.mu = mu

    def get_data(self) -> List[Tuple[np.ndarray, InterpolationIndices]]:
        r"""
        Return the interpolation data for the strongly monotone operator.

        :returns: A list containing one tuple with the interpolation matrix and indices.
        :rtype: List[Tuple[np.ndarray, InterpolationIndices]]
        """
        matrix = 0.5 * np.array([
            [2 * self.mu, -2 * self.mu, -1, 1],
            [-2 * self.mu, 2 * self.mu,  1, -1],
            [-1, 1, 0, 0],
            [1, -1, 0, 0]
        ])
        interp_idx = InterpolationIndices("i<j")
        return [(matrix, interp_idx)]

class LipschitzOperator(OperatorInterpolationCondition):
    r"""
    Operator interpolation condition for a Lipschitz operator.

    :param L: The Lipschitz parameter (must be > 0 and finite).
    :type L: float
    :raises ValueError: If L is not a number, <= 0, or infinite.
    """
    def __init__(self, L: Union[int, float]):
        if not isinstance(L, (int, float)):
            raise ValueError("Lipschitz parameter must be a number.")
        L = float(L)
        if L <= 0 or L == float('inf'):
            raise ValueError("Lipschitz parameter (L) must be greater than 0 and finite.")
        self.L = L

    def get_data(self) -> List[Tuple[np.ndarray, InterpolationIndices]]:
        r"""
        Return the interpolation data for the Lipschitz operator.

        :returns: A list containing one tuple with the interpolation matrix and indices.
        :rtype: List[Tuple[np.ndarray, InterpolationIndices]]
        """
        matrix = np.array([
            [-self.L**2, self.L**2, 0, 0],
            [self.L**2, -self.L**2, 0, 0],
            [0, 0, 1, -1],
            [0, 0, -1, 1]
        ])
        interp_idx = InterpolationIndices("i<j")
        return [(matrix, interp_idx)]

class Cocoercive(OperatorInterpolationCondition):
    r"""
    Operator interpolation condition for a cocoercive operator.

    :param beta: The cocoercivity parameter (must be > 0 and finite).
    :type beta: float
    :raises ValueError: If beta is not a number, <= 0, or infinite.
    """
    def __init__(self, beta: Union[int, float]):
        if not isinstance(beta, (int, float)):
            raise ValueError("Cocoercivity parameter must be a number.")
        beta = float(beta)
        if beta <= 0 or beta == float('inf'):
            raise ValueError("Cocoercivity parameter (beta) must be greater than 0 and finite.")
        self.beta = beta

    def get_data(self) -> List[Tuple[np.ndarray, InterpolationIndices]]:
        r"""
        Return the interpolation data for the cocoercive operator.

        :returns: A list containing one tuple with the interpolation matrix and indices.
        :rtype: List[Tuple[np.ndarray, InterpolationIndices]]
        """
        matrix = 0.5 * np.array([
            [0, 0, -1, 1],
            [0, 0,  1, -1],
            [-1, 1, 2 * self.beta, -2 * self.beta],
            [1, -1, -2 * self.beta, 2 * self.beta]
        ])
        interp_idx = InterpolationIndices("i<j")
        return [(matrix, interp_idx)]

class WeakMintyVariationalInequality(OperatorInterpolationCondition):
    r"""
    Operator interpolation condition for an operator that fulfills the weak Minty variational inequality.

    :param rho_minty: The weak MVI parameter (must be >= 0 and finite).
    :type rho_minty: float
    :raises ValueError: If rho_minty is not a number, negative, or infinite.
    """
    def __init__(self, rho_minty: Union[int, float]):
        if not isinstance(rho_minty, (int, float)):
            raise ValueError("Weak MVI parameter must be a number.")
        rho_minty = float(rho_minty)
        if rho_minty < 0 or rho_minty == float('inf'):
            raise ValueError("Weak MVI parameter (rho_minty) must be nonnegative and finite.")
        self.rho_minty = rho_minty

    def get_data(self) -> List[Tuple[np.ndarray, InterpolationIndices]]:
        r"""
        Return the interpolation data for the weak Minty variational inequality condition.

        :returns: A list containing one tuple with the interpolation matrix and indices.
        :rtype: List[Tuple[np.ndarray, InterpolationIndices]]
        """
        matrix = 0.5 * np.array([
            [0, 0, -1, 0],
            [0, 0,  1, 0],
            [-1, 1, -self.rho_minty, 0],
            [0, 0, 0, 0]
        ])
        interp_idx = InterpolationIndices("i!=star")
        return [(matrix, interp_idx)]


# ---------------------------------------------------------------------------
# Concrete Function Interpolation Condition Classes
# ---------------------------------------------------------------------------
class ParametrizedFunctionInterpolationCondition(FunctionInterpolationCondition):
    r"""
    Base class for function interpolation conditions that depend on parameters mu and L.

    Provides a helper function to compute the interpolation data based on mu and L.
    This base class is used for both smooth and nonsmooth conditions by setting L appropriately.

    :param mu: The convexity parameter. For strongly convex functions, mu > 0; for convex functions, mu = 0;
               for weakly convex functions, mu < 0.
    :type mu: float
    :param L: The smoothness parameter. For nonsmooth functions, L is set to infinity.
    :type L: float
    :raises ValueError: If mu is not a number or if mu is not less than L.
    """
    def __init__(self, mu: Union[int, float], L: Union[int, float]):
        if not isinstance(mu, (int, float)):
            raise ValueError("Parameter mu must be a number.")
        if not isinstance(L, (int, float)):
            raise ValueError("Parameter L must be a number.")
        mu = float(mu)
        L = float(L)
        if mu == float('-inf'):
            raise ValueError("ParametrizedFunctionInterpolationCondition: mu cannot be -inf.")
        if not (mu < L):
            raise ValueError("ParametrizedFunctionInterpolationCondition requires that -inf < mu < L <= +inf.")
        self.mu = mu
        self.L = L

    @staticmethod
    def _compute_interpolation_data(mu: float, L: float) -> Tuple[np.ndarray, np.ndarray, bool, InterpolationIndices]:
        r"""
        Compute the interpolation data based on mu and L.

        When L is infinite, the condition is nonsmooth and a simpler interpolation matrix is used.
        When L is finite, the interpolation data follows the formula for smooth functions.

        :param mu: The convexity parameter.
        :type mu: float
        :param L: The smoothness parameter.
        :type L: float
        :returns: A tuple containing the matrix, vector, eq flag, and interpolation indices.
        :rtype: Tuple[np.ndarray, np.ndarray, bool, InterpolationIndices]
        """
        if L == float('inf'):
            matrix = 0.5 * np.array([
                [mu, -mu, 0, 1],
                [-mu, mu, 0, -1],
                [0, 0, 0, 0],
                [1, -1, 0, 0]
            ])
        else:
            matrix = (1 / (2 * (L - mu))) * np.array([
                [L * mu, -L * mu, -mu, L],
                [-L * mu, L * mu, mu, -L],
                [-mu, mu, 1, -1],
                [L, -L, -1, 1]
            ])
        vector = np.array([-1, 1])
        eq = False
        interp_idx = InterpolationIndices("i!=j")
        return matrix, vector, eq, interp_idx

    def get_data(self) -> List[Tuple[np.ndarray, np.ndarray, bool, InterpolationIndices]]:
        r"""
        Return the interpolation data for the function condition.

        :returns: A list containing one tuple with the matrix, vector, eq flag, and interpolation indices.
        :rtype: List[Tuple[np.ndarray, np.ndarray, bool, InterpolationIndices]]
        """
        return [self._compute_interpolation_data(self.mu, self.L)]

class Convex(ParametrizedFunctionInterpolationCondition):
    r"""
    Function interpolation condition for a convex function.

    This condition uses mu = 0 and L = infinity.
    """
    def __init__(self):
        super().__init__(mu=0.0, L=float('inf'))

class StronglyConvex(ParametrizedFunctionInterpolationCondition):
    r"""
    Function interpolation condition for a strongly convex function.

    :param mu: Strong convexity parameter (must be > 0 and finite).
    :type mu: float
    L is set to infinity.
    :raises ValueError: If mu is not valid.
    """
    def __init__(self, mu: Union[int, float]):
        if not isinstance(mu, (int, float)):
            raise ValueError("Parameter mu must be a number.")
        mu = float(mu)
        if mu <= 0:
            raise ValueError("For StronglyConvex, mu must be > 0.")
        if mu == float('inf'):
            raise ValueError("For StronglyConvex, mu must be finite.")
        super().__init__(mu=mu, L=float('inf'))

class WeaklyConvex(ParametrizedFunctionInterpolationCondition):
    r"""
    Function interpolation condition for a weakly convex function.

    :param mu_tilde: Weak convexity parameter (must be > 0).
    :type mu_tilde: float
    L is set to infinity.
    :raises ValueError: If mu is not valid.
    """
    def __init__(self, mu_tilde: Union[int, float]):
        mu = - mu_tilde
        if not isinstance(mu, (int, float)):
            raise ValueError("Parameter mu must be a number.")
        mu = float(mu)
        if mu >= 0:
            raise ValueError("For WeaklyConvex, mu_tilde must be > 0.")
        super().__init__(mu=mu, L=float('inf'))

class Smooth(ParametrizedFunctionInterpolationCondition):
    r"""
    Function interpolation condition for a smooth function.

    :param L: Smoothness parameter (must be > 0 and finite).
    :type L: float
    This condition uses mu = -L.
    :raises ValueError: If L is not valid.
    """
    def __init__(self, L: Union[int, float]):
        if not isinstance(L, (int, float)):
            raise ValueError("Parameter L must be a number.")
        L = float(L)
        if L <= 0 or L == float('inf'):
            raise ValueError("For Smooth, L must be > 0 and finite.")
        super().__init__(mu=-L, L=L)

class SmoothConvex(ParametrizedFunctionInterpolationCondition):
    r"""
    Function interpolation condition for a smooth convex function.

    :param L: Smoothness parameter (must be > 0 and finite).
    :type L: float
    This condition uses mu = 0.
    :raises ValueError: If L is not valid.
    """
    def __init__(self, L: Union[int, float]):
        if not isinstance(L, (int, float)):
            raise ValueError("Parameter L must be a number.")
        L = float(L)
        if L <= 0 or L == float('inf'):
            raise ValueError("For SmoothConvex, L must be > 0 and finite.")
        super().__init__(mu=0.0, L=L)

class SmoothStronglyConvex(ParametrizedFunctionInterpolationCondition):
    r"""
    Function interpolation condition for a smooth strongly convex function.

    :param mu: Strong convexity parameter (must be > 0 and finite).
    :type mu: float
    :param L: Smoothness parameter (must be > 0 and finite) with mu < L.
    :type L: float
    :raises ValueError: If parameters are not valid.
    """
    def __init__(self, mu: Union[int, float], L: Union[int, float]):
        if not isinstance(mu, (int, float)):
            raise ValueError("Parameter mu must be a number.")
        if not isinstance(L, (int, float)):
            raise ValueError("Parameter L must be a number.")
        mu = float(mu)
        L = float(L)
        if mu <= 0:
            raise ValueError("For SmoothStronglyConvex, mu must be > 0.")
        if mu == float('inf'):
            raise ValueError("For SmoothStronglyConvex, mu must be finite.")
        if L <= 0 or L == float('inf'):
            raise ValueError("For SmoothStronglyConvex, L must be > 0 and finite.")
        if mu >= L:
            raise ValueError("For SmoothStronglyConvex, mu must be less than L.")
        super().__init__(mu=mu, L=L)

class SmoothWeaklyConvex(ParametrizedFunctionInterpolationCondition):
    r"""
    Function interpolation condition for a smooth weakly convex function.

    :param mu_tilde: Weak convexity parameter (must be > 0).
    :type mu_tilde: float
    :param L: Smoothness parameter (must be > 0 and finite).
    :type L: float
    :raises ValueError: If parameters are not valid.
    """
    def __init__(self, mu_tilde: Union[int, float], L: Union[int, float]):
        mu = - mu_tilde
        if not isinstance(mu, (int, float)):
            raise ValueError("Parameter mu_tilde must be a number.")
        if not isinstance(L, (int, float)):
            raise ValueError("Parameter L must be a number.")
        mu = float(mu)
        L = float(L)
        if mu >= 0:
            raise ValueError("For SmoothWeaklyConvex, mu_tilde must be > 0.")
        if L <= 0 or L == float('inf'):
            raise ValueError("For SmoothWeaklyConvex, L must be > 0 and finite.")
        super().__init__(mu=mu, L=L)

class IndicatorFunctionOfClosedConvexSet(FunctionInterpolationCondition):
    r"""
    Function interpolation condition for the indicator function of a closed convex set.

    This condition does not require any parameters.
    """
    def get_data(self) -> List[Tuple[np.ndarray, np.ndarray, bool, InterpolationIndices]]:
        r"""
        Return the interpolation data for the indicator function.

        :returns: A list containing two tuples with the interpolation data.
        :rtype: List[Tuple[np.ndarray, np.ndarray, bool, InterpolationIndices]]
        """
        interp_idx_ineq = InterpolationIndices("i!=j")
        matrix_ineq = 0.5 * np.array([
            [0, 0, 0, 1],
            [0, 0, 0, -1],
            [0, 0, 0, 0],
            [1, -1, 0, 0]
        ])
        vector_ineq = np.array([0, 0])
        interp_idx_eq = InterpolationIndices("i")
        matrix_eq = np.array([[0, 0], [0, 0]])
        vector_eq = np.array([1])
        return [
            (matrix_ineq, vector_ineq, False, interp_idx_ineq),
            (matrix_eq, vector_eq, True, interp_idx_eq)
        ]

class SupportFunctionOfClosedConvexSet(FunctionInterpolationCondition):
    r"""
    Function interpolation condition for the support function of a closed convex set.

    This condition does not require any parameters.
    """
    def get_data(self) -> List[Tuple[np.ndarray, np.ndarray, bool, InterpolationIndices]]:
        r"""
        Return the interpolation data for the support function.

        :returns: A list containing two tuples with the interpolation data.
        :rtype: List[Tuple[np.ndarray, np.ndarray, bool, InterpolationIndices]]
        """
        interp_idx_ineq = InterpolationIndices("i!=j")
        matrix_ineq = 0.5 * np.array([
            [0, 0, 0, 0],
            [0, 0, 1, -1],
            [0, 1, 0, 0],
            [0, -1, 0, 0]
        ])
        vector_ineq = np.array([0, 0])
        interp_idx_eq = InterpolationIndices("i")
        matrix_eq = 0.5 * np.array([[0, 1], [1, 0]])
        vector_eq = np.array([-1])
        return [
            (matrix_ineq, vector_ineq, False, interp_idx_ineq),
            (matrix_eq, vector_eq, True, interp_idx_eq)
        ]

class GradientDominated(FunctionInterpolationCondition):
    r"""
    Function interpolation condition for gradient-dominated functions.

    :param mu_gd: The gradient-dominated parameter (must be > 0 and finite).
    :type mu_gd: float
    :raises ValueError: If mu_gd is not a number, <= 0, or infinite.
    """
    def __init__(self, mu_gd: Union[int, float]):
        if not isinstance(mu_gd, (int, float)):
            raise ValueError("Gradient-dominated parameter must be a number.")
        mu_gd = float(mu_gd)
        if mu_gd <= 0 or mu_gd == float('inf'):
            raise ValueError("Gradient-dominated parameter (mu_gd) must be greater than 0 and finite.")
        self.mu_gd = mu_gd

    def get_data(self) -> List[Tuple[np.ndarray, np.ndarray, bool, InterpolationIndices]]:
        r"""
        Return the interpolation data for gradient-dominated functions.

        :returns: A list containing two tuples with the interpolation data.
        :rtype: List[Tuple[np.ndarray, np.ndarray, bool, InterpolationIndices]]
        """
        a1 = np.array([-1, 1])
        M1 = np.zeros((4, 4))
        
        a2 = np.array([1, -1])
        M2 = np.zeros((4, 4))
        M2[2, 2] = -1 / (2 * self.mu_gd)
        
        interp_idx = InterpolationIndices("i!=star")
        eq_flag = False
        return [
            (M1, a1, eq_flag, interp_idx),
            (M2, a2, eq_flag, interp_idx)
        ]


# ---------------------------------------------------------------------------
# InclusionProblem Class
# ---------------------------------------------------------------------------
class InclusionProblem:
    r"""
    Encapsulates an inclusion problem of the form:

    .. math::
       \text{find } y \in H \text{ such that } 0 \in \sum_{i \in I_{\text{func}}} \partial f_i(y) + \sum_{i \in I_{\text{op}}} G_i(y)

    :param components: A list where each element represents a component (1-indexed).
                       Each element is either a single interpolation condition instance or a list of such instances.
                       All conditions for a given component must be of the same type (either all operator conditions or all function conditions).
    :type components: List[Union[InterpolationCondition, List[InterpolationCondition]]]
    :raises ValueError: If components is empty or if components contain invalid conditions.
    """
    def __init__(self, components: List[Union[InterpolationCondition, List[InterpolationCondition]]]):
        if not components or len(components) < 1:
            raise ValueError("Error in InclusionProblem __init__: At least one component is required.")
        self.m = len(components)
        self.components: Dict[int, List[InterpolationCondition]] = {}
        # Convert the list to a dictionary with 1-indexed keys.
        for i, comp in enumerate(components, start=1):
            if isinstance(comp, list):
                for v in comp:
                    if not isinstance(v, InterpolationCondition):
                        raise ValueError(
                            f"Error in InclusionProblem __init__: Component {i} must contain only "
                            f"InterpolationCondition instances. Got {type(v)}."
                        )
                self.components[i] = comp
            else:
                if not isinstance(comp, InterpolationCondition):
                    raise ValueError(
                        f"Error in InclusionProblem __init__: Component {i} must be an "
                        f"InterpolationCondition instance. Got {type(comp)}."
                    )
                self.components[i] = [comp]
            self._validate_component_uniformity(i, self.components[i])
            self._validate_component_data(i, self.components[i])
        
        for conditions in self.components.values():
            if any(isinstance(cond, GradientDominated) for cond in conditions):
                if self.m != 1:
                    raise ValueError(
                        "Error: If any component contains a GradientDominated instance, "
                        "the total number of components (m) must be exactly 1."
                    )

        for conditions in self.components.values():
            if any(isinstance(cond, WeakMintyVariationalInequality) for cond in conditions):
                if self.m != 1:
                    raise ValueError(
                        "Error: If any component contains a WeakMintyVariationalInequality instance, "
                        "the total number of components (m) must be exactly 1."
                    )

        self.I_op = [k for k, conds in self.components.items() 
                     if isinstance(conds[0], OperatorInterpolationCondition)]
        self.I_func = [k for k, conds in self.components.items() 
                       if isinstance(conds[0], FunctionInterpolationCondition)]
    
    def _validate_component_uniformity(self, index: int, conditions: List[InterpolationCondition]) -> None:
        r"""
        Ensure that all conditions for a given component are of the same type.

        :param index: The component index.
        :type index: int
        :param conditions: A list of interpolation conditions.
        :type conditions: List[InterpolationCondition]
        :raises ValueError: If the conditions contain a mix of operator and function conditions.
        """
        is_operator = [isinstance(v, OperatorInterpolationCondition) for v in conditions]
        is_function = [isinstance(v, FunctionInterpolationCondition) for v in conditions]
        if any(is_operator) and any(is_function):
            raise ValueError(
                f"Error: Component {index} contains a mix of operator and function "
                "interpolation conditions."
            )
    
    def _validate_condition_data(self, cond: InterpolationCondition) -> None:
        r"""
        Validate the data returned by an interpolation condition.

        :param cond: An interpolation condition instance.
        :type cond: InterpolationCondition
        :raises ValueError: If the condition data does not conform to the expected structure.
        """
        data = cond.get_data()
        if isinstance(cond, OperatorInterpolationCondition):
            for tup in data:
                if not (isinstance(tup, tuple) and len(tup) == 2):
                    raise ValueError(
                        f"Error: Operator condition data must be a tuple of 2 elements. Received: {tup}"
                    )
                matrix, interp_idx = tup
                if not isinstance(matrix, np.ndarray):
                    raise ValueError("Error: Operator interpolation matrix must be a numpy array.")
                if matrix.ndim != 2 or matrix.shape[0] != matrix.shape[1]:
                    raise ValueError("Error: Operator interpolation matrix must be square.")
                if not np.allclose(matrix, matrix.T, atol=1e-8):
                    raise ValueError("Error: Operator interpolation matrix must be symmetric.")
                if not isinstance(interp_idx, InterpolationIndices):
                    raise ValueError("Error: Operator interpolation indices must be an instance of InterpolationIndices.")
        elif isinstance(cond, FunctionInterpolationCondition):
            for tup in data:
                if not (isinstance(tup, tuple) and len(tup) == 4):
                    raise ValueError(
                        f"Error: Function condition data must be a tuple of 4 elements. Received: {tup}"
                    )
                matrix, vector, eq, interp_idx = tup
                if not isinstance(matrix, np.ndarray):
                    raise ValueError("Error: Function interpolation matrix must be a numpy array.")
                if matrix.ndim != 2 or matrix.shape[0] != matrix.shape[1]:
                    raise ValueError("Error: Function interpolation matrix must be square.")
                if not np.allclose(matrix, matrix.T, atol=1e-8):
                    raise ValueError("Error: Function interpolation matrix must be symmetric.")
                if not isinstance(vector, np.ndarray):
                    raise ValueError("Error: Function interpolation vector must be a numpy array.")
                if vector.ndim != 1:
                    raise ValueError("Error: Function interpolation vector must be 1-dimensional.")
                if matrix.shape[0] != 2 * vector.shape[0]:
                    raise ValueError(
                        "Error: Function interpolation matrix rows must equal 2 times the length of the vector."
                    )
                if not isinstance(interp_idx, InterpolationIndices):
                    raise ValueError("Error: Function interpolation indices must be an instance of InterpolationIndices.")
        else:
            raise ValueError("Error: Unknown interpolation condition type.")
    
    def _validate_component_data(self, index: int, conditions: List[InterpolationCondition]) -> None:
        r"""
        Validate the data of all conditions for a given component.

        :param index: The component index.
        :type index: int
        :param conditions: A list of interpolation conditions.
        :type conditions: List[InterpolationCondition]
        """
        for cond in conditions:
            self._validate_condition_data(cond)
    
    def get_component_data(self, index: int) -> List[Union[
        Tuple[np.ndarray, InterpolationIndices],
        Tuple[np.ndarray, np.ndarray, bool, InterpolationIndices]
    ]]:
        r"""
        Retrieve the raw interpolation data for a given component.

        :param index: The component index (1-indexed).
        :type index: int
        :returns: A list of tuples containing the interpolation data. For operator conditions, each tuple is
                  ``(matrix, interpolation_indices)``; for function conditions, each tuple is 
                  ``(matrix, vector, eq, interpolation_indices)``.
        :rtype: List[Union[Tuple[np.ndarray, InterpolationIndices], Tuple[np.ndarray, np.ndarray, bool, InterpolationIndices]]]
        :raises ValueError: If the index is not defined.
        """
        if index < 1 or index > self.m:
            raise ValueError(
                f"Error: Index must be in 1,...,{self.m}. Component {index} is not defined."
            )
        data = []
        for cond in self.components[index]:
            data.extend(cond.get_data())
        return data
    
    def get_component(self, index: int) -> List[InterpolationCondition]:
        r"""
        Retrieve the interpolation condition instances for a given component.

        :param index: The component index (1-indexed).
        :type index: int
        :returns: A list of interpolation condition instances.
        :rtype: List[InterpolationCondition]
        :raises ValueError: If the index is not defined.
        """
        if index < 1 or index > self.m:
            raise ValueError(
                f"Error: Index must be in 1,...,{self.m}. Component {index} is not defined."
            )
        return self.components[index]
    
    def update_component_instances(self, 
                                   index: int, 
                                   new_instances: Union[InterpolationCondition, List[InterpolationCondition]]):
        r"""
        Update the interpolation condition instances for a given component.

        This replaces the full list of instances for that component.

        :param index: The component index (1-indexed).
        :type index: int
        :param new_instances: A single interpolation condition instance or a list of them.
        :type new_instances: Union[InterpolationCondition, List[InterpolationCondition]]
        :raises ValueError: If the index is not defined or if new_instances are invalid.
        """
        if index < 1 or index > self.m:
            raise ValueError(
                f"Error: Index must be in 1,...,{self.m}. Component {index} is not defined."
            )
        if isinstance(new_instances, list):
            for inst in new_instances:
                if not isinstance(inst, InterpolationCondition):
                    raise ValueError(
                        f"Error: All new instances for component {index} must be "
                        "InterpolationCondition instances."
                    )
            op_flags = [isinstance(inst, OperatorInterpolationCondition) for inst in new_instances]
            func_flags = [isinstance(inst, FunctionInterpolationCondition) for inst in new_instances]
            if any(op_flags) and any(func_flags):
                raise ValueError(
                    f"Error: New instances for component {index} must be all operator "
                    "or all function conditions; mixing is not allowed."
                )
            self.components[index] = new_instances
        else:
            if not isinstance(new_instances, InterpolationCondition):
                raise ValueError(
                    f"Error: New instance for component {index} must be an "
                    "InterpolationCondition instance."
                )
            self.components[index] = [new_instances]
        self._validate_component_uniformity(index, self.components[index])
        self._validate_component_data(index, self.components[index])



from .helper_functions import create_symmetric_matrix_expression
from .helper_functions import create_symmetric_matrix

__all__ = [
    'create_symmetric_matrix_expression',
    'create_symmetric_matrix',
]



import numpy as np
import math
from typing import Tuple
from .algorithm import Algorithm

class ITEM(Algorithm):
    def __init__(self, mu, L):
        super().__init__(2, 1, [1], [1], [])
        self.mu = mu
        self.L = L
    
    def set_L(self, L):
        self.L = L

    def set_mu(self, mu):
        self.mu = mu
    
    def get_A(self, k: int) -> float:
        q = self.mu / self.L
        A = 0.0 
        for _ in range(k):
            A = ((1 + q) * A + 2 * (1 + math.sqrt((1 + A) * (1 + q * A)))) / ((1 - q) ** 2)
        return A

    def compute_beta(self, k: int) -> float:
        q = self.mu / self.L
        A_k = self.get_A(k)
        A_k1 = self.get_A(k + 1)
        return A_k / ((1 - q) * A_k1)

    def compute_delta(self, k: int) -> float:
        q = self.mu / self.L
        A_k = self.get_A(k)
        A_k1 = self.get_A(k + 1)
        numerator = ((1 - q) ** 2) * A_k1 - (1 + q) * A_k
        denominator = 2 * (1 + q + q * A_k)
        return numerator / denominator

    def get_ABCD(self, k: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        q = self.mu / self.L
        A_k = self.get_A(k)
        A_k1 = self.get_A(k + 1)
        beta = self.compute_beta(k)
        delta = self.compute_delta(k)

        A = np.array([[beta, 1-beta],
                      [q*beta*delta, 1-q*beta*delta]])
        
        B = np.array([[-1/self.L],
                      [-delta/self.L]])
        
        C = np.array([[beta, 1-beta]])
        
        D = np.array([[0]])
        
        return (A, B, C, D)


import numpy as np
from typing import Tuple
from .algorithm import Algorithm

class NesterovFastGradientMethod(Algorithm):
    def __init__(self, gamma):
        super().__init__(2, 1, [2], [1], [])
        self.gamma = gamma
    
    def set_gamma(self, gamma):
        self.gamma = gamma
    
    def get_ABCD(self, k: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        lambda_var = 1
        for _ in range(0, k + 1):
            lambda_var_prev = lambda_var
            lambda_var = (1 + np.sqrt(1 + 4 * lambda_var ** 2)) / 2

        alpha = (lambda_var_prev - 1) / lambda_var

        A = np.array([[1+alpha, -alpha],
                      [1, 0]])
        
        B = np.array([[-self.gamma, 0],
                      [0, 0]])
        
        C = np.array([[1+alpha, -alpha],
                      [1, 0]])
        
        D = np.array([[0, 0],
                      [0, 0]])
        
        return (A, B, C, D)



import numpy as np
from typing import Tuple
from .algorithm import Algorithm

class ProxSkip(Algorithm):
    # https://arxiv.org/pdf/2202.09357
    def __init__(self, gamma, L):
        self.set_gamma(gamma)  
        self.set_L(L) 
        super().__init__(2, 2, [self.L, 1], [1, 2], [])

    def set_gamma(self, gamma):
        if not isinstance(gamma, (int, float)) or gamma <= 0:
            raise ValueError("gamma must be a real number greater than 0")
        self.gamma = gamma

    def set_L(self, L):
        if not isinstance(L, int) or L < 1:
            raise ValueError("L must be an integer greater than or equal to 1")
        self.L = L
    
    def get_ABCD(self, k: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        A = np.array([[1, self.gamma*(1-self.L)],
                      [0, 0]])
        B = np.vstack((np.hstack((-self.gamma * np.ones((1,self.L)), np.array([[  -self.gamma * (2*self.L - 1) ]]))),
                       np.hstack((np.zeros((1, self.L)), np.array([[-1]])))))
        C = np.vstack((np.array([[1, 0]]),
                       np.hstack((np.ones((self.L, 1)), self.gamma * (1 - self.L) * np.ones((self.L, 1))))))
        D = self.create_D()
        return (A, B, C, D)

    def create_D(self):
        D = np.zeros((self.L+1, self.L+1))
        for i in range(1, self.L):
            D[i, :i] = -self.gamma  
            D[i, -1] = -self.gamma*( i - 1 + self.L)
        D[-1, 0] = -self.gamma  
        D[-1, -1] = -self.gamma * self.L
        return D


import numpy as np
from typing import Tuple
from .algorithm import Algorithm

class NesterovConstant(Algorithm):
    # https://link.springer.com/book/10.1007/978-3-319-91578-4 
    # See Chapter 2.2: Constant step scheme, III
    def __init__(self, sigma, beta):
        super().__init__(2, 1, [1], [1], [])
        self.sigma = sigma
        self.beta = beta
    
    def set_sigma(self, sigma):
        self.sigma = sigma

    def set_beta(self, beta):
        self.beta = beta
    
    def get_ABCD(self, k: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        q = self.sigma/self.beta
        
        A = np.array([[2/(1+np.sqrt(q)),-(1-np.sqrt(q))/(1+np.sqrt(q))],[1,0]])
        B = np.array([[-1/beta],[0]])
        C = np.array([[2/(1+np.sqrt(q)),-(1-np.sqrt(q))/(1+np.sqrt(q))]])
        D = np.array([[0]])
        return (A, B, C, D)


import numpy as np
from typing import Tuple
from .algorithm import Algorithm

class MalitskyTamFRB(Algorithm):
    def __init__(self, gamma):
        super().__init__(2, 2, [2, 1], [], [1, 2])
        self.gamma = gamma
    
    def set_gamma(self, gamma):
        self.gamma = gamma
    
    def get_ABCD(self, k: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        A = np.array([[1, 0],
                      [1, 0]])
        B = np.array([[-2*self.gamma, self.gamma, -self.gamma],
                      [0, 0, 0]])
        C = np.array([[1, 0],
                      [0, 1],
                      [1, 0]])
        D = np.array([[0, 0, 0], 
                      [0, 0, 0],
                      [-2*self.gamma, self.gamma, -self.gamma]])
        return (A, B, C, D)


import numpy as np
from typing import Tuple
from .algorithm import Algorithm

class AcceleratedProximalPoint(Algorithm):
    def __init__(self, gamma, type: str = "operator"):
        if type == "operator":
            super().__init__(3, 1, [1], [], [1])
        elif type == "function":
            super().__init__(3, 1, [1], [1], [])
        else:
            raise ValueError("type must be either 'operator' or 'function'")
        self.gamma = gamma
    
    def set_gamma(self, gamma):
        self.gamma = gamma
    
    def get_ABCD(self, k: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        lambda_var = k / (k + 2)
        A = np.array([[0, 1 , 0], 
                      [-2*lambda_var, 1+lambda_var, lambda_var],
                      [0, 1, 0]])
        B = np.array([[-self.gamma],
                      [-self.gamma*(1+lambda_var)],
                      [0]])
        C = np.array([[0, 1, 0]])
        D = np.array([[-self.gamma]])
        return (A, B, C, D)


import numpy as np
from typing import Type, Optional, Tuple, Union, List
from itertools import combinations, product
from mosek.fusion import Model, Domain, OptimizeError
import mosek.fusion.pythonic
from autolyap.utils.helper_functions import create_symmetric_matrix_expression
from autolyap.problemclass import InclusionProblem
from autolyap.algorithms import Algorithm

class LinearConvergence:
    @staticmethod
    def get_parameters_distance_to_solution(
            algo: Type[Algorithm], 
            h: int = 0, 
            alpha: int = 0,
            i: int = 1, 
            j: int = 1, 
            tau: int = 0
        ) -> Union[Tuple[np.ndarray, np.ndarray],
                   Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
                   ]:
        r"""
        Compute the matrices for the distance to solution.

        This method computes the following matrix

        .. math::
            P = \left( P_{(i,j)}\, Y_\tau^{0,h} - P_{(i,\star)}\, Y_\star^{0,h} \right)^T
                \left( P_{(i,j)}\, Y_\tau^{0,h} - P_{(i,\star)}\, Y_\star^{0,h} \right),

        and constructs :math:`T` (and, if functional components exist, the vectors :math:`p` and :math:`t`) as zero.

        **Definitions**

        - :math:`Y_\tau^{0,h}` is the Y matrix at iteration :math:`\tau` over the horizon :math:`[0, h]`.
        - :math:`Y_\star^{0,h}` is the “star” Y matrix over :math:`[0, h]`.
        - :math:`P_{(i,j)}` and :math:`P_{(i,\star)}` are the projection matrices for component :math:`i`.

        **Dimensions**

        - :math:`\dim(P) = n + (h+1)\bar{m} + m,`
        - :math:`\dim(T) = n + (h+\alpha+2)\bar{m} + m,`
        - If :math:`algo.m_func > 0`:
        
          - :math:`\mathrm{len}(p) = (h+1)\bar{m}_{\text{func}} + m\_func,`
          - :math:`\mathrm{len}(t) = (h+\alpha+2)\bar{m}_{\text{func}} + m\_func.`

        **Parameters**
        
        :param algo: An instance of Algorithm. It must provide:
                   - ``algo.m`` (total number of components),
                   - ``algo.m_bar_is`` (a list where the *i*-th entry gives the number of evaluations for component *i*),
                   - Methods ``get_Ys(k_min, k_max)`` and ``get_Ps()``.
        :param h: A nonnegative integer defining the time horizon :math:`[0, h]` for Y matrices.
        :param alpha: A nonnegative integer for extending the horizon for :math:`T` (and :math:`t`).
        :param i: Component index (1-indexed). Default is 1; must satisfy :math:`1 \le i \le algo.m`.
        :param j: Evaluation index for component :math:`i`. Default is 1; must satisfy :math:`1 \le j \le algo.m_bar_is[i-1]`.
        :param tau: Iteration index. Default is 0; must satisfy :math:`0 \le \tau \le h`.

        **Returns**
        
        - If :math:`algo.m_func == 0`:
          
          :return: A tuple ``(P, T)``
        
        - Otherwise:
        
          :return: A tuple ``(P, p, T, t)``, where :math:`P` is computed as above and :math:`T`, :math:`p`, and :math:`t` are zero arrays with the appropriate dimensions.

        **Raises**
        
        :raises ValueError: If any input is out of its valid range or if required matrices are missing.
        """
        # ----- Input Checking -----
        if not isinstance(h, int) or h < 0:
            raise ValueError("Parameter h must be a nonnegative integer.")
        if not isinstance(alpha, int) or alpha < 0:
            raise ValueError("Parameter alpha must be a nonnegative integer.")
        
        if not isinstance(i, int) or i < 1 or i > algo.m:
            raise ValueError(f"Component index i must be in [1, {algo.m}]. Got {i}.")
        
        num_eval = algo.m_bar_is[i - 1]
        if not isinstance(j, int) or j < 1 or j > num_eval:
            raise ValueError(f"For component {i}, evaluation index j must be in [1, {num_eval}]. Got {j}.")
        
        if not isinstance(tau, int) or tau < 0 or tau > h:
            raise ValueError(f"Iteration index tau must be in [0, {h}]. Got {tau}.")

        # ----- Dimensions for P and T -----
        n = algo.n            # State dimension.
        m = algo.m            # Total number of components.
        m_bar = algo.m_bar    # Total evaluations per iteration.
        
        # Dimension of P: n + (h+1)*m_bar + m.
        dim_P = n + (h + 1) * m_bar + m
        
        # Dimension of T: n + (h+alpha+2)*m_bar + m.
        dim_T = n + (h + alpha + 2) * m_bar + m

        # ----- Compute P (nonzero) -----
        # Retrieve Y matrices for the horizon [0, h].
        Ys = algo.get_Ys(0, h)
        if tau not in Ys:
            raise ValueError(f"Y matrix for iteration tau = {tau} not found.")
        if 'star' not in Ys:
            raise ValueError("Y star matrix ('star') not found.")
        
        # Retrieve projection matrices.
        Ps = algo.get_Ps()
        if (i, j) not in Ps:
            raise ValueError(f"Projection matrix for component {i}, evaluation {j} not found.")
        if (i, 'star') not in Ps:
            raise ValueError(f"Projection matrix for component {i} star not found.")
        
        # Compute the difference:
        diff = Ps[(i, j)] @ Ys[tau] - Ps[(i, 'star')] @ Ys['star']
        # Compute the outer product:
        P_mat = diff.T @ diff
        
        # ----- Construct T, p, and t as zeros with appropriate dimensions -----
        T_mat = np.zeros((dim_T, dim_T))
        if algo.m_func > 0:
            m_bar_func = algo.m_bar_func    # Total evaluations for functional components.
            m_func = algo.m_func            # Number of functional components.
            dim_p = (h + 1) * m_bar_func + m_func
            p_vec = np.zeros(dim_p)
            dim_t = (h + alpha + 2) * m_bar_func + m_func
            t_vec = np.zeros(dim_t)
            return P_mat, p_vec, T_mat, t_vec
        else:
            return P_mat, T_mat

    @staticmethod
    def get_parameters_function_value_suboptimality(
            algo: Type[Algorithm],
            h: int = 0,
            alpha: int = 0,
            j: int = 1,
            tau: int = 0
        ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        r"""
        Compute the matrices/vectors for function-value suboptimality.

        This function is only applicable when :math:`m = m\_func = 1`.

        It returns a tuple :math:`(P, p, T, t)` where:

        - :math:`p` is computed as

          .. math::
              p = \left( F_{(1,j,tau)}^{0,h} - F_{(1,\star,\star)}^{0,h} \right)^T,

          with :math:`p` returned as a 1D numpy array of length 
          :math:`(h+1)\bar{m}_{\text{func}} + m\_func`.
        - :math:`P` is a zero matrix of dimension

          .. math::
              n + (h+1)\bar{m} + m,
          
        - :math:`T` is a zero matrix of dimension

          .. math::
              n + (h+\alpha+2)\bar{m} + m,
          
        - :math:`t` is a zero vector of dimension

          .. math::
              (h+\alpha+2)\bar{m}_{\text{func}} + m\_func.

        **Parameters**

        :param algo: An instance of Algorithm. It must satisfy :math:`algo.m = 1` and :math:`algo.m\_func = 1`.
        :param h: A nonnegative integer defining the horizon :math:`[0, h]` for F matrices.
        :param alpha: A nonnegative integer for extending the horizon for :math:`T` and :math:`t`.
        :param j: Evaluation index for component 1. Default is 1; must satisfy :math:`1 \le j \le algo.m_bar_is[0]`.
        :param tau: Iteration index. Default is 0; must satisfy :math:`0 \le \tau \le h`.

        **Returns**

        :return: A tuple :math:`(P, p, T, t)`, where :math:`p` is computed as above (a 1D numpy array)
                 and :math:`P`, :math:`T`, and :math:`t` are zero arrays with appropriate dimensions.

        **Raises**

        :raises ValueError: If :math:`algo.m \ne 1` or :math:`algo.m\_func \ne 1`, if any input parameter is out of range,
                            or if the required F matrices are not found.
        """
        # ----- Check that m and m_func equal 1 -----
        if algo.m != 1 or algo.m_func != 1:
            raise ValueError("get_parameters_function_value_suboptimality is only applicable when m = m_func = 1.")
        
        # ----- Validate inputs -----
        if not isinstance(h, int) or h < 0:
            raise ValueError("Parameter h must be a nonnegative integer.")
        if not isinstance(alpha, int) or alpha < 0:
            raise ValueError("Parameter alpha must be a nonnegative integer.")
        
        num_eval = algo.m_bar_is[0]
        if not isinstance(j, int) or j < 1 or j > num_eval:
            raise ValueError(f"For component 1, evaluation index j must be in [1, {num_eval}]. Got {j}.")
        
        if not isinstance(tau, int) or tau < 0 or tau > h:
            raise ValueError(f"Iteration index tau must be in [0, {h}]. Got {tau}.")
        
        # ----- Retrieve F matrices for the horizon [0, h] -----
        Fs = algo.get_Fs(0, h)
        key_nonstar = (1, j, tau)
        key_star = (1, 'star', 'star')
        if key_nonstar not in Fs:
            raise ValueError(f"F matrix for key {key_nonstar} not found.")
        if key_star not in Fs:
            raise ValueError("F star matrix (1, 'star', 'star') not found.")
        
        # Compute p as the difference between F matrices, then convert to 1D.
        p_vec = (Fs[key_nonstar] - Fs[key_star]).T
        p_vec = np.ravel(p_vec)  # Ensure p is a 1D numpy array.
        
        # ----- Determine dimensions -----
        n = algo.n                    # State dimension.
        m = algo.m                    # Total number of components (should be 1).
        m_bar = algo.m_bar            # Total evaluations per iteration.
        m_bar_func = algo.m_bar_func  # Evaluations for functional components.
        m_func = algo.m_func          # Number of functional components (should be 1).
        
        dim_P = n + (h + 1) * m_bar + m
        dim_T = n + (h + alpha + 2) * m_bar + m
        dim_p = (h + 1) * m_bar_func + m_func
        dim_t = (h + alpha + 2) * m_bar_func + m_func
        
        # ----- Construct zero matrices/vectors for the remaining outputs -----
        P_mat = np.zeros((dim_P, dim_P))
        T_mat = np.zeros((dim_T, dim_T))
        t_vec = np.zeros(dim_t)
        
        return P_mat, p_vec, T_mat, t_vec
    
    @staticmethod
    def bisection_search_rho(
            prob: Type[InclusionProblem],
            algo: Type[Algorithm],
            P: np.ndarray,
            T: np.ndarray,
            p: Optional[np.ndarray] = None,
            t: Optional[np.ndarray] = None,
            h: int = 0,
            alpha: int = 0,
            Q_equals_P: bool = False,
            S_equals_T: bool = False,
            q_equals_p: bool = False,
            s_equals_t: bool = False,
            remove_C2: bool = False,
            remove_C3: bool = False,
            remove_C4: bool = True,
            lower_bound: float = 0.0,
            upper_bound: float = 1.0,
            tol: float = 1e-12
        ) -> Optional[float]:
        r"""
        Perform a bisection search to find the minimal contraction parameter :math:`\rho`.

        This method performs a bisection search over :math:`\rho` in the interval 
        :math:`[{\text{lower\_bound}}, {\text{upper\_bound}}]` to find the minimal value for which the 
        iteration-independent Lyapunov inequality holds. At each iteration it calls
        ``verify_iteration_independent_Lyapunov`` until the interval size is below :math:`{\text{tol}}`.

        **Parameters**

        :param prob: An InclusionProblem instance containing interpolation conditions.
        :param algo: An Algorithm instance providing dimensions and methods.
        :param P: A symmetric matrix of dimension 
                  :math:`n + (h+1)\bar{m} + m` by :math:`n + (h+1)\bar{m} + m`.
        :param T: A symmetric matrix of dimension 
                  :math:`n + (h+\alpha+2)\bar{m} + m` by :math:`n + (h+\alpha+2)\bar{m} + m`.
        :param p: A vector for functional components (if applicable).
        :param t: A vector for functional components (if applicable).
        :param h: Nonnegative integer defining the history for the matrices.
        :param alpha: Nonnegative integer for extending the horizon.
        :param Q_equals_P: If True, set Q equal to P.
        :param S_equals_T: If True, set S equal to T.
        :param q_equals_p: For functional components, if True, set q equal to p.
        :param s_equals_t: For functional components, if True, set s equal to t.
        :param remove_C2: Flag to remove constraint C2.
        :param remove_C3: Flag to remove constraint C3.
        :param remove_C4: Flag to remove constraint C4.
        :param lower_bound: Lower bound for :math:`\rho`.
        :param upper_bound: Upper bound for :math:`\rho`.
        :param tol: Tolerance for the bisection search stopping criterion.

        **Returns**

        :return: The minimal :math:`\rho` in :math:`[{\text{lower\_bound}}, {\text{upper\_bound}}]` that verifies the 
                 Lyapunov inequality within tolerance :math:`{\text{tol}}`, or ``None`` if the inequality does not hold 
                 at the upper bound.
        """
        # Ensure that the inequality holds at the initial upper bound.
        if not IterationIndependent.verify_iteration_independent_Lyapunov(
                prob, algo, P, T, p, t, rho=upper_bound, h=h, alpha=alpha,
                Q_equals_P=Q_equals_P, S_equals_T=S_equals_T,
                q_equals_p=q_equals_p, s_equals_t=s_equals_t,
                remove_C2=remove_C2, remove_C3=remove_C3, remove_C4=remove_C4):
            return None

        l = lower_bound
        u = upper_bound
        while (u - l) > tol:
            mid = (l + u) / 2.0
            if IterationIndependent.verify_iteration_independent_Lyapunov(
                    prob, algo, P, T, p, t, rho=mid, h=h, alpha=alpha,
                    Q_equals_P=Q_equals_P, S_equals_T=S_equals_T,
                    q_equals_p=q_equals_p, s_equals_t=s_equals_t,
                    remove_C2=remove_C2, remove_C3=remove_C3, remove_C4=remove_C4):
                u = mid
            else:
                l = mid

        return u

class SublinearConvergence:
    @staticmethod
    def get_parameters_fixed_point_residual(
            algo: Type[Algorithm],
            h: int = 0,
            alpha: int = 0,
            tau: int = 0
        ) -> Union[Tuple[np.ndarray, np.ndarray],
                   Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
                   ]:
        r"""
        Compute the matrices for the fixed-point residual.

        For a given iteration index :math:`\tau` (with :math:`0 \le \tau \le h+\alpha+1`), define

        .. math::
            T = \left( X_{\tau+1}^{0, h+\alpha+1} - X_{\tau}^{0, h+\alpha+1} \right)^T
                \left( X_{\tau+1}^{0, h+\alpha+1} - X_{\tau}^{0, h+\alpha+1} \right),

        where :math:`X_{\tau}^{0, h+\alpha+1}` is the X matrix computed over the horizon :math:`[0, h+\alpha+1]` 
        (via ``algo.get_Xs``).

        **Dimensions**

        - :math:`P` is a zero matrix of dimension :math:`n + (h+1)\bar{m} + m`.
        - :math:`T` is computed as above and has dimension :math:`n + (h+\alpha+2)\bar{m} + m`.
        - If :math:`algo.m_func > 0`:
        
          - :math:`p` is a zero vector of length :math:`(h+1)\bar{m}_{\text{func}} + m\_func,`
          - :math:`t` is a zero vector of length :math:`(h+\alpha+2)\bar{m}_{\text{func}} + m\_func.`

        **Parameters**

        :param algo: An instance of Algorithm.
        :param h: A nonnegative integer defining the time horizon :math:`[0, h]` for :math:`P`.
        :param alpha: A nonnegative integer for extending the horizon for :math:`T` (and :math:`t`).
        :param tau: Iteration index for computing the fixed-point residual.
                  Must satisfy :math:`0 \le \tau \le h+\alpha+1`.

        **Returns**

        - If :math:`algo.m_func == 0`:
          
          :return: A tuple ``(P, T)``.
        
        - Otherwise:
          
          :return: A tuple ``(P, p, T, t)``.

        **Raises**

        :raises ValueError: If any input parameter is out of its valid range or if the required X matrices are missing.
        """
        # ----- Input Checking -----
        if not isinstance(h, int) or h < 0:
            raise ValueError("Parameter h must be a nonnegative integer.")
        if not isinstance(alpha, int) or alpha < 0:
            raise ValueError("Parameter alpha must be a nonnegative integer.")
        if not isinstance(tau, int) or tau < 0 or tau > h + alpha + 1:
            raise ValueError(f"Iteration index tau must be in [0, {h+alpha+1}]. Got {tau}.")

        # ----- Dimensions for P and T -----
        n = algo.n         # State dimension.
        m = algo.m         # Total number of components.
        m_bar = algo.m_bar # Total evaluations per iteration.
        
        # Dimension of P: n + (h+1)*m_bar + m.
        dim_P = n + (h + 1) * m_bar + m
        # Dimension of T: n + (h+alpha+2)*m_bar + m.
        dim_T = n + (h + alpha + 2) * m_bar + m

        # ----- Compute T -----
        # Retrieve X matrices for the horizon [0, h+alpha+1].
        # Note: get_Xs returns X_tau for tau in [0, (h+alpha+1)+1] = [0, h+alpha+2].
        Xs = algo.get_Xs(0, h + alpha + 1)
        if tau not in Xs or (tau + 1) not in Xs:
            raise ValueError(f"X matrices for iterations tau = {tau} and tau+1 = {tau+1} not found.")
        
        diff = Xs[tau + 1] - Xs[tau]
        T_mat = diff.T @ diff

        # ----- Construct P, p, and t as zeros with appropriate dimensions -----
        P_mat = np.zeros((dim_P, dim_P))
        if algo.m_func > 0:
            m_bar_func = algo.m_bar_func    # Evaluations for functional components.
            m_func = algo.m_func            # Number of functional components.
            dim_p = (h + 1) * m_bar_func + m_func
            p_vec = np.zeros(dim_p)
            dim_t = (h + alpha + 2) * m_bar_func + m_func
            t_vec = np.zeros(dim_t)
            return P_mat, p_vec, T_mat, t_vec
        else:
            return P_mat, T_mat

    @staticmethod
    def get_parameters_duality_gap(
            algo: Type[Algorithm],
            h: int = 0,
            alpha: int = 0,
            tau: int = 0
        ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        r"""
        Compute the matrices for the duality gap.

        For a given iteration index :math:`\tau` (with :math:`0 \le \tau \le h+\alpha+1`), define

        .. math::
            T = -\frac{1}{2} \sum_{i=1}^{m} \begin{bmatrix}
            P_{(i,\star)}\, U_\star^{0,h+\alpha+1} \\
            P_{(i,1)}\, Y_\tau^{0,h+\alpha+1}
            \end{bmatrix}^T
            \begin{bmatrix}
            0 & 1 \\
            1 & 0 
            \end{bmatrix}
            \begin{bmatrix}
            P_{(i,\star)}\, U_\star^{0,h+\alpha+1} \\
            P_{(i,1)}\, Y_\tau^{0,h+\alpha+1}
            \end{bmatrix},

        and

        .. math::
            t = \sum_{i=1}^{m} \left( F_{(i,1,\tau)}^{0,h+\alpha+1} - F_{(i,\star,\star)}^{0,h+\alpha+1} \right)^T.

        All other matrices are set to zero.

        **Requirements**

        It is required that :math:`m = m\_func` (i.e. all components are functional).

        **Dimensions**

        - :math:`P` is a zero matrix of dimension :math:`n + (h+1)\bar{m} + m`.
        - :math:`p` is a zero vector of length :math:`(h+1)\bar{m}_{\text{func}} + m\_func`.
        - :math:`T` is computed as above and has dimension :math:`n + (h+\alpha+2)\bar{m} + m`.
        - :math:`t` is computed as above and has length :math:`(h+\alpha+2)\bar{m}_{\text{func}} + m\_func`.

        **Parameters**

        :param algo: An instance of Algorithm (with :math:`m = m\_func`).
        :param h: A nonnegative integer defining the time horizon :math:`[0, h]` for :math:`P`.
        :param alpha: A nonnegative integer for extending the horizon for :math:`T` and :math:`t`.
        :param tau: Iteration index for computing the duality gap.
                  Must satisfy :math:`0 \le \tau \le h+\alpha+1`.

        **Returns**

        :return: A tuple :math:`(P, p, T, t)`, where :math:`t` is a one-dimensional array.

        **Raises**

        :raises ValueError: If any input parameter is out of its valid range, if the required matrices
                            are missing, or if :math:`m \ne m\_func`.
        """
        # ----- Check that m = m_func -----
        if algo.m != algo.m_func:
            raise ValueError("get_parameters_duality_gap is only applicable when m = m_func.")
        
        # ----- Input Checking -----
        if not isinstance(h, int) or h < 0:
            raise ValueError("Parameter h must be a nonnegative integer.")
        if not isinstance(alpha, int) or alpha < 0:
            raise ValueError("Parameter alpha must be a nonnegative integer.")
        if not isinstance(tau, int) or tau < 0 or tau > h + alpha + 1:
            raise ValueError(f"Iteration index tau must be in [0, {h+alpha+1}]. Got {tau}.")
        
        # ----- Dimensions for P, T, p, and t -----
        n = algo.n         # State dimension.
        m = algo.m         # Total number of components (also equals m_func here).
        m_bar = algo.m_bar # Total evaluations per iteration.
        
        # Dimension of P: n + (h+1)*m_bar + m.
        dim_P = n + (h + 1) * m_bar + m
        # Dimension of T: n + (h+alpha+2)*m_bar + m.
        dim_T = n + (h + alpha + 2) * m_bar + m
        
        # Functional dimensions:
        m_bar_func = algo.m_bar_func
        m_func = algo.m_func
        dim_p = (h + 1) * m_bar_func + m_func
        dim_t = (h + alpha + 2) * m_bar_func + m_func
        
        # ----- Compute T -----
        # Retrieve U and Y matrices over the horizon [0, h+alpha+1]
        U_dict = algo.get_Us(0, h + alpha + 1)
        Y_dict = algo.get_Ys(0, h + alpha + 1)
        if 'star' not in U_dict:
            raise ValueError("U_star matrix ('star') not found.")
        if tau not in Y_dict:
            raise ValueError(f"Y matrix for iteration tau = {tau} not found.")
        U_star = U_dict['star']
        Y_tau = Y_dict[tau]
        
        # Retrieve projection matrices.
        Ps = algo.get_Ps()
        
        # Define the 2x2 swap matrix (renamed to mid).
        mid = np.array([[0, 1],
                        [1, 0]])
        
        # Initialize the accumulator for T.
        T_sum = np.zeros((dim_T, dim_T))
        for i in range(1, m + 1):
            # Retrieve P_{(i,star)} and P_{(i,1)}.
            if (i, 'star') not in Ps:
                raise ValueError(f"Projection matrix for component {i} star not found.")
            if (i, 1) not in Ps:
                raise ValueError(f"Projection matrix for component {i}, evaluation 1 not found.")
            P_i_star = Ps[(i, 'star')]
            P_i_1 = Ps[(i, 1)]
            
            # Compute the two blocks.
            block1 = P_i_star @ U_star  # 1 x dim_T
            block2 = P_i_1 @ Y_tau        # 1 x dim_T
            
            # Stack to form a 2 x dim_T matrix.
            block = np.vstack([block1, block2])
            # Contribution from component i: block^T mid block.
            T_sum += block.T @ mid @ block
        T_mat = -0.5 * T_sum
        
        # ----- Compute t -----
        # Retrieve F matrices over the horizon [0, h+alpha+1].
        Fs = algo.get_Fs(0, h + alpha + 1)
        t_sum = np.zeros((dim_t, 1))
        for i in range(1, m + 1):
            key_nonstar = (i, 1, tau)
            key_star = (i, 'star', 'star')
            if key_nonstar not in Fs:
                raise ValueError(f"F matrix for key {key_nonstar} not found.")
            if key_star not in Fs:
                raise ValueError(f"F star matrix for key {key_star} not found.")
            diff_F = Fs[key_nonstar] - Fs[key_star]  # This is a row vector.
            t_sum += diff_F.T  # Sum the transposed (column) vectors.
        # Flatten to obtain a one-dimensional array.
        t_vec = t_sum.ravel()
        
        # ----- Construct P and p as zeros -----
        P_mat = np.zeros((dim_P, dim_P))
        p_vec = np.zeros(dim_p)
        
        return P_mat, p_vec, T_mat, t_vec
    
    @staticmethod
    def get_parameters_function_value_suboptimality(
            algo: Type[Algorithm],
            h: int = 0,
            alpha: int = 0,
            j: int = 1,
            tau: int = 0
        ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        r"""
        Compute the matrices/vectors for function value suboptimality.

        This function is only applicable when :math:`m = m\_func = 1`.

        It returns a tuple :math:`(P, p, T, t)` where:

        - :math:`t` is computed as

          .. math::
              t = \left( F_{(1,j,\tau)}^{0,h+\alpha+1} - F_{(1,\star,\star)}^{0,h+\alpha+1} \right)^T,

          with :math:`t` returned as a 1D numpy array of length 
          :math:`(h+\alpha+2)\bar{m}_{\text{func}} + m\_func`.
        - :math:`P` is a zero matrix of dimension

          .. math::
              n + (h+1)\bar{m} + m,
          
        - :math:`T` is a zero matrix of dimension

          .. math::
              n + (h+\alpha+2)\bar{m} + m,
          
        - :math:`t` is a zero vector of dimension

          .. math::
              (h+1)\bar{m}_{\text{func}} + m\_func.

        **Parameters**

        :param algo: An instance of Algorithm. It must satisfy :math:`algo.m = 1` and :math:`algo.m\_func = 1`.
        :param h: A nonnegative integer defining the horizon :math:`[0, h]` for F matrices.
        :param alpha: A nonnegative integer for extending the horizon for :math:`T` and :math:`t`.
        :param j: Evaluation index for component 1. Default is 1; must satisfy :math:`1 \le j \le algo.m_bar_is[0]`.
        :param tau: Iteration index. Default is 0; must satisfy :math:`0 \le \tau \le h+\alpha+1`.

        **Returns**

        :return: A tuple :math:`(P, p, T, t)`, where :math:`t` is computed as above (a 1D numpy array)
                 and :math:`P`, :math:`T`, and :math:`p` are zero arrays with appropriate dimensions.

        **Raises**

        :raises ValueError: If :math:`algo.m \ne 1` or :math:`algo.m\_func \ne 1`, if any input parameter is out of range,
                            or if the required F matrices are not found.
        """
        # ----- Check that m and m_func equal 1 -----
        if algo.m != 1 or algo.m_func != 1:
            raise ValueError("get_parameters_function_value_suboptimality is only applicable when m = m_func = 1.")
        
        # ----- Validate inputs -----
        if not isinstance(h, int) or h < 0:
            raise ValueError("Parameter h must be a nonnegative integer.")
        if not isinstance(alpha, int) or alpha < 0:
            raise ValueError("Parameter alpha must be a nonnegative integer.")
        
        num_eval = algo.m_bar_is[0]
        if not isinstance(j, int) or j < 1 or j > num_eval:
            raise ValueError(f"For component 1, evaluation index j must be in [1, {num_eval}]. Got {j}.")
        
        if not isinstance(tau, int) or tau < 0 or tau > h:
            raise ValueError(f"Iteration index tau must be in [0, {h+alpha+1}]. Got {tau}.")
        
        # ----- Dimensions for P, T, p, and t -----
        n = algo.n         # State dimension.
        m = algo.m         # Total number of components (also equals m_func here).
        m_bar = algo.m_bar # Total evaluations per iteration.
        
        # Dimension of P: n + (h+1)*m_bar + m.
        dim_P = n + (h + 1) * m_bar + m
        # Dimension of T: n + (h+alpha+2)*m_bar + m.
        dim_T = n + (h + alpha + 2) * m_bar + m

        # Functional dimensions:
        m_bar_func = algo.m_bar_func
        m_func = algo.m_func
        dim_p = (h + 1) * m_bar_func + m_func
        dim_t = (h + alpha + 2) * m_bar_func + m_func

        T = np.zeros((dim_T, dim_T))
        P = np.zeros((dim_P, dim_P))
        p = np.zeros(dim_p)
        
        # ----- Compute t -----
        # Retrieve F matrices over the horizon [0, h+alpha+1].
        Fs = algo.get_Fs(0, h + alpha + 1)
        key_nonstar = (1, j, tau)
        key_star = (1, 'star', 'star')
        if key_nonstar not in Fs:
            raise ValueError(f"F matrix for key {key_nonstar} not found.")
        if key_star not in Fs:
            raise ValueError(f"F star matrix for key {key_star} not found.")
        t = Fs[key_nonstar] - Fs[key_star]  # This is a row vector.
        # Flatten to obtain a one-dimensional array.
        t = t.ravel()

        return P, p, T, t
        
    @staticmethod
    def get_parameters_optimality_measure(
                algo: Type[Algorithm],
                h: int = 0,
                alpha: int = 0,
                tau: int = 0
        ) -> Union[Tuple[np.ndarray, np.ndarray],
                   Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
                   ]:
        r"""
        Compute the matrices for the optimality measure.

        For a given iteration index :math:`\tau` (with :math:`0 \le \tau \le h+\alpha+1`), define

        .. math::
            T =
            \begin{cases}
              \left( P_{(1,1)}\, U_\tau^{0,h+\alpha+1} \right)^T \left( P_{(1,1)}\, U_\tau^{0,h+\alpha+1} \right)
              & \text{if } m = 1, \\[1em]
              \left( \left( \sum_{i=1}^{m} P_{(i,1)}\, U_\tau^{0,h+\alpha+1} \right)^T \left( \sum_{i=1}^{m} P_{(i,1)}\, U_\tau^{0,h+\alpha+1} \right)
              + \sum_{i=2}^{m} \left( \left( P_{(1,1)} - P_{(i,1)} \right) Y_\tau^{0,h+\alpha+1} \right)^T \left( \left( P_{(1,1)} - P_{(i,1)} \right) Y_\tau^{0,h+\alpha+1} \right) \right)
              & \text{if } m > 1.
            \end{cases}

        All other matrices are set to zero.

        **Dimensions**

        - :math:`P` is a zero matrix of dimension :math:`n + (h+1)\bar{m} + m`.
        - If :math:`algo.m_func > 0`:
        
          - :math:`p` is a zero vector of length :math:`(h+1)\bar{m}_{\text{func}} + m\_func,`
          - :math:`t` is a zero vector of length :math:`(h+\alpha+2)\bar{m}_{\text{func}} + m\_func.`

        **Parameters**

        :param algo: An instance of Algorithm.
        :param h: A nonnegative integer defining the time horizon :math:`[0, h]` for :math:`P`.
        :param alpha: A nonnegative integer for extending the horizon for :math:`T` (and :math:`t`).
        :param tau: Iteration index for computing the optimality measure.
                  Must satisfy :math:`0 \le \tau \le h+\alpha+1`.

        **Returns**

        - If :math:`algo.m_func == 0`:
          
          :return: A tuple ``(P, T)``.
        
        - Otherwise:
          
          :return: A tuple ``(P, p, T, t)``.

        **Raises**

        :raises ValueError: If any input parameter is out of its valid range or if required matrices are missing.
        """
        # ----- Input Checking -----
        if not isinstance(h, int) or h < 0:
            raise ValueError("Parameter h must be a nonnegative integer.")
        if not isinstance(alpha, int) or alpha < 0:
            raise ValueError("Parameter alpha must be a nonnegative integer.")
        if not isinstance(tau, int) or tau < 0 or tau > h + alpha + 1:
            raise ValueError(f"Iteration index tau must be in [0, {h+alpha+1}]. Got {tau}.")
        
        # ----- Dimensions for P and T -----
        n = algo.n         # State dimension.
        m = algo.m         # Total number of components.
        m_bar = algo.m_bar # Total evaluations per iteration.
        
        # Dimension of P: n + (h+1)*m_bar + m.
        dim_P = n + (h + 1) * m_bar + m
        # Dimension of T: n + (h+alpha+2)*m_bar + m.
        dim_T = n + (h + alpha + 2) * m_bar + m
        
        # ----- Retrieve U and Y matrices over [0, h+alpha+1] -----
        U_dict = algo.get_Us(0, h + alpha + 1)
        Y_dict = algo.get_Ys(0, h + alpha + 1)
        if tau not in U_dict:
            raise ValueError(f"U matrix for iteration tau = {tau} not found.")
        if tau not in Y_dict:
            raise ValueError(f"Y matrix for iteration tau = {tau} not found.")
        U_tau = U_dict[tau]
        Y_tau = Y_dict[tau]
        
        # ----- Retrieve projection matrices -----
        Ps = algo.get_Ps()
        
        # ----- Compute T -----
        if m == 1:
            # Case: m = 1.
            if (1, 1) not in Ps:
                raise ValueError("Projection matrix for component 1, evaluation 1 not found.")
            P_11 = Ps[(1, 1)]
            block = P_11 @ U_tau
            T_mat = block.T @ block
        else:
            # Case: m > 1.
            # First term: sum_{i=1}^{m} P_{(i,1)} U_tau.
            sum_U = None
            for i in range(1, m + 1):
                if (i, 1) not in Ps:
                    raise ValueError(f"Projection matrix for component {i}, evaluation 1 not found.")
                P_i1 = Ps[(i, 1)]
                term = P_i1 @ U_tau
                sum_U = term if sum_U is None else sum_U + term
            first_term = sum_U.T @ sum_U
            
            # Second term: sum_{i=2}^{m} ((P_{(1,1)} - P_{(i,1)}) Y_k).T ((P_{(1,1)} - P_{(i,1)}) Y_k).
            if (1, 1) not in Ps:
                raise ValueError("Projection matrix for component 1, evaluation 1 not found.")
            P_11 = Ps[(1, 1)]
            second_term = np.zeros((dim_T, dim_T))
            for i in range(2, m + 1):
                if (i, 1) not in Ps:
                    raise ValueError(f"Projection matrix for component {i}, evaluation 1 not found.")
                diff = (P_11 - Ps[(i, 1)]) @ Y_tau
                second_term += diff.T @ diff
            T_mat = first_term + second_term
        
        # ----- Construct zero matrices for the remaining outputs -----
        P_mat = np.zeros((dim_P, dim_P))
        
        if algo.m_func > 0:
            m_bar_func = algo.m_bar_func    # Evaluations for functional components.
            m_func = algo.m_func            # Number of functional components.
            dim_p = (h + 1) * m_bar_func + m_func
            dim_t = (h + alpha + 2) * m_bar_func + m_func
            p_vec = np.zeros(dim_p)
            t_vec = np.zeros(dim_t)
            return P_mat, p_vec, T_mat, t_vec
        else:
            return P_mat, T_mat


class IterationIndependent:
    
    LinearConvergence = LinearConvergence
    SublinearConvergence = SublinearConvergence

    @staticmethod
    def verify_iteration_independent_Lyapunov(
            prob: Type[InclusionProblem],
            algo: Type[Algorithm],
            P: np.ndarray,
            T: np.ndarray,
            p: Optional[np.ndarray] = None,
            t: Optional[np.ndarray] = None,
            rho: float = 1.0,
            h: int = 0,
            alpha: int = 0,
            Q_equals_P: bool = False,
            S_equals_T: bool = False,
            q_equals_p: bool = False,
            s_equals_t: bool = False,
            remove_C2: bool = False,
            remove_C3: bool = False,
            remove_C4: bool = True
        ) -> bool:
        r"""
        Verify an iteration-independent Lyapunov inequality via an SDP.

        This method sets up and solves a semidefinite program (SDP) using Mosek Fusion to 
        verify a Lyapunov inequality for a given inclusion problem and algorithm.

        **Parameters**

        :param prob: An InclusionProblem instance containing interpolation conditions.
        :param algo: An Algorithm instance providing dimensions and methods to compute matrices.
        :param P: A symmetric numpy array of dimension 
                  :math:`n + (h+1)\bar{m} + m` by :math:`n + (h+1)\bar{m} + m`.
        :param T: A symmetric numpy array of dimension 
                  :math:`n + (h+\alpha+2)\bar{m} + m` by :math:`n + (h+\alpha+2)\bar{m} + m`.
        :param p: A vector for functional components (if any) with appropriate dimensions.
        :param t: A vector for functional components (if any) with appropriate dimensions.
        :param rho: A scalar contraction parameter used in forming the Lyapunov inequality.
        :param h: Nonnegative integer defining history.
        :param alpha: Nonnegative integer defining overlap.
        :param Q_equals_P: If True, sets Q equal to P.
        :param S_equals_T: If True, sets S equal to T.
        :param q_equals_p: For functional components, if True, sets q equal to p.
        :param s_equals_t: For functional components, if True, sets s equal to t.
        :param remove_C2: Flag to remove constraint C2.
        :param remove_C3: Flag to remove constraint C3.
        :param remove_C4: Flag to remove constraint C4.

        **Returns**

        :return: True if the SDP is solved successfully (i.e. a Lyapunov inequality exists), False otherwise.

        **Raises**

        :raises ValueError: If input dimensions or other conditions are violated.
        """

        # -------------------------------------------------------------------------
        # Validate consistency between the problem and the algorithm.
        # -------------------------------------------------------------------------
        if prob.m != algo.m:
            raise ValueError("Mismatch in number of components: prob.m and algo.m must be the same")
        
        # Check that the functional and operator component indices are identical.
        if set(prob.I_func) != set(algo.I_func):
            raise ValueError("Mismatch in functional component indices between prob and algo")
        if set(prob.I_op) != set(algo.I_op):
            raise ValueError("Mismatch in operator component indices between prob and algo")
        
        # Ensure h and alpha are nonnegative.
        if h < 0:
            raise ValueError("Parameter 'h' must be >= 0.")
        if alpha < 0:
            raise ValueError("Parameter 'alpha' must be >= 0.")
        
        # -------------------------------------------------------------------------
        # Retrieve dimensions from the algorithm instance.
        # -------------------------------------------------------------------------
        n = algo.n                      # State dimension.
        m_bar = algo.m_bar              # Total evaluations per iteration.
        m = algo.m                      # Total number of components.
        m_bar_func = algo.m_bar_func    # Total evaluations for functional components.
        m_func = algo.m_func            # Number of functional components.
        m_op = algo.m_op                # Number of operator components.
        m_bar_op = algo.m_bar_op        # Total evaluations for operator components.
        
        # Expected dimension for matrix P: [n + (h+1)*m_bar + m] x [n + (h+1)*m_bar + m].
        dim_P = n + (h + 1) * m_bar + m
        if not (isinstance(P, np.ndarray) and P.ndim == 2 and P.shape[0] == P.shape[1] == dim_P):
            raise ValueError(f"P must be a symmetric matrix of dimension {dim_P}x{dim_P}. Got shape {P.shape}.")
        if not np.allclose(P, P.T, atol=1e-8):
            raise ValueError("P must be symmetric.")
        
        # Expected dimension for matrix T: [n + (h+alpha+2)*m_bar + m] x [n + (h+alpha+2)*m_bar + m].
        dim_T = n + (h + alpha + 2) * m_bar + m
        if not (isinstance(T, np.ndarray) and T.ndim == 2 and T.shape[0] == T.shape[1] == dim_T):
            raise ValueError(f"T must be a symmetric matrix of dimension {dim_T}x{dim_T}. Got shape {T.shape}.")
        if not np.allclose(T, T.T, atol=1e-8):
            raise ValueError("T must be symmetric.")
        
        # For functional components, p and t must have proper dimensions.
        if m_func > 0:
            # Compute required dimensions
            dim_p = (h + 1) * m_bar_func + m_func
            dim_t = (h + alpha + 2) * m_bar_func + m_func

            # Check p
            if p is None:
                raise ValueError(
                    f"p must be a 1D numpy array of length {dim_p}, but got None."
                )
            if not (isinstance(p, np.ndarray) and p.ndim == 1 and p.shape[0] == dim_p):
                raise ValueError(
                    f"p must be a 1D numpy array of length {dim_p}. Got shape "
                    f"{getattr(p, 'shape', None)}."
                )

            # Check t
            if t is None:
                raise ValueError(
                    f"t must be a 1D numpy array of length {dim_t}, but got None."
                )
            if not (isinstance(t, np.ndarray) and t.ndim == 1 and t.shape[0] == dim_t):
                raise ValueError(
                    f"t must be a 1D numpy array of length {dim_t}. Got shape "
                    f"{getattr(t, 'shape', None)}."
                )


        # -------------------------------------------------------------------------
        # Set up the Mosek Fusion model and decision variables.
        # -------------------------------------------------------------------------
        with Model() as Mod:
            # Q variable: either set equal to P or defined as a new symmetric variable.
            if Q_equals_P:
                Q = P
            else:
                Qij = Mod.variable("Q_upper_triangle_vars", dim_P * (dim_P + 1) // 2, Domain.unbounded())
                Q = create_symmetric_matrix_expression(Qij, dim_P)

            # S variable: either set equal to T or defined as a new symmetric variable.
            if S_equals_T:
                S = T
            else:
                Sij = Mod.variable("S_upper_triangle_vars", dim_T * (dim_T + 1) // 2, Domain.unbounded())
                S = create_symmetric_matrix_expression(Sij, dim_T)
            
            # For functional components, create variables q and s (or set them equal to p and t).
            if m_func > 0:
                if q_equals_p:
                    q = p
                else:
                    q = Mod.variable("q", dim_p, Domain.unbounded())
                if s_equals_t:
                    s = t
                else:
                    s = Mod.variable("s", dim_t, Domain.unbounded())
            
            # ---------------------------------------------------------------------
            # Build the main PSD (positive semidefinite) and equality constraint sums.
            # These will later be constrained to be in the PSD cone or equal zero, respectively.
            # ---------------------------------------------------------------------
            Ws = {}       # Dictionary for matrix constraint sums.
            k_maxs = {}   # Dictionary to store the maximum iteration index for each condition.

            # For condition "C1": use _compute_Thetas with k_max = h + alpha + 1.
            Theta0_C1, Theta1_C1 = IterationIndependent._compute_Thetas(algo, h, alpha, condition='C1')
            Ws["C1"] = Theta1_C1.T @ Q @ Theta1_C1 - rho * Theta0_C1.T @ Q @ Theta0_C1 + S
            k_maxs["C1"] = h + alpha + 1

            # Condition "C2": if not removed, enforce P - Q.
            if not remove_C2:
                Ws["C2"] = P - Q
                k_maxs["C2"] = h

            # Condition "C3": if not removed, enforce T - S.
            if not remove_C3:
                Ws["C3"] = T - S
                k_maxs["C3"] = h + alpha + 1

            # Condition "C4": if not removed, use _compute_Thetas with k_max = h + alpha + 2.
            if not remove_C4:
                Theta0_C4, Theta1_C4 = IterationIndependent._compute_Thetas(algo, h, alpha, condition='C4')
                Ws["C4"] = Theta1_C4.T @ S @ Theta1_C4 - Theta0_C4.T @ S @ Theta0_C4
                k_maxs["C4"] = h + alpha + 2

            # For functional components, build the analogous vector constraints.
            if m_func > 0:
                ws = {}
                theta0_C1, theta1_C1 = IterationIndependent._compute_thetas(algo, h, alpha, condition='C1')
                ws["C1"] = (theta1_C1.T - rho * theta0_C1.T) @ q + s

                if not remove_C2:
                    ws["C2"] = p - q

                if not remove_C3:
                    ws["C3"] = t - s

                if not remove_C4:
                    theta0_C4, theta1_C4 = IterationIndependent._compute_thetas(algo, h, alpha, condition='C4')
                    ws["C4"] = (theta1_C4.T - theta0_C4.T) @ s

            # Initialize lists of active conditions.
            conds = ["C1"]
            if not remove_C2:
                conds.append("C2")
            if not remove_C3:
                conds.append("C3")
            if not remove_C4:
                conds.append("C4")

            # Initialize dictionaries to sum up the PSD and equality constraints.
            PSD_constraint_sums = {}
            eq_constraint_sums = {}
            for cond in conds:
                PSD_constraint_sums[cond] = -Ws[cond]
                if m_func > 0:
                    eq_constraint_sums[cond] = -ws[cond]

            # Dictionaries to hold multipliers for interpolation conditions.
            if m_op > 0:
                lambdas_op = {}
            if m_func > 0:
                lambdas_func = {}
                nus_func = {}

            # ---------------------------------------------------------------------
            # Define inner helper functions for processing interpolation data.
            # These functions handle both operator and function conditions.
            # ---------------------------------------------------------------------
            def process_pairs(cond: str, 
                            i: int, 
                            o: int, 
                            interpolation_data: Union[Tuple[np.ndarray, str],
                                                        Tuple[np.ndarray, np.ndarray, bool, str]], 
                            pairs: List[Union[Tuple[int, int], Tuple[str, str]]], 
                            comp_type: str) -> None:
                r"""
                Process a given set of interpolation pairs.

                Constructs a unique key based on condition, component, pairs, and interpolation index.
                For operator conditions, creates a nonnegative multiplier.
                For functional conditions, if the constraint is an equality, creates an unrestricted multiplier;
                otherwise, creates a nonnegative multiplier.
                Updates the PSD and equality constraint sums accordingly.

                :param cond: The condition identifier (e.g. "C1", "C2", etc.).
                :param i: The component index.
                :param o: The interpolation index.
                :param interpolation_data: Interpolation data tuple.
                :param pairs: A list of (j, k) pairs.
                :param comp_type: A string, either 'op' for operator or 'func' for functional.
                """
                key_parts = [cond, f"i:{i}"]
                for (j, k) in pairs:
                    key_parts.append(f"(j:{j},k:{k})")
                key_parts.append(f"o:{o}")
                key = tuple(key_parts)

                if comp_type == 'op':
                    M, _ = interpolation_data
                else:
                    M, a, eq, _ = interpolation_data

                # Compute the lifted matrix W for the given pairs.
                W_matrix = algo.compute_W(i, pairs, 0, k_maxs[cond], M)

                if comp_type == 'op':
                    lambda_var = Mod.variable(1, Domain.greaterThan(0.0))
                    lambdas_op[key] = lambda_var
                    PSD_constraint_sums[cond] = PSD_constraint_sums[cond] + lambda_var[0] * W_matrix
                else:
                    # For functional components, compute the aggregated F vector.
                    F_vector = algo.compute_F_aggregated(i, pairs, 0, k_maxs[cond], a)
                    if eq:
                        nu_var = Mod.variable(1, Domain.unbounded())
                        nus_func[key] = nu_var
                        PSD_constraint_sums[cond] = PSD_constraint_sums[cond] + nu_var[0] * W_matrix
                        eq_constraint_sums[cond] = eq_constraint_sums[cond] + nu_var[0] * F_vector
                    else:
                        lambda_var = Mod.variable(1, Domain.greaterThan(0.0))
                        lambdas_func[key] = lambda_var
                        PSD_constraint_sums[cond] = PSD_constraint_sums[cond] + lambda_var[0] * W_matrix
                        eq_constraint_sums[cond] = eq_constraint_sums[cond] + lambda_var[0] * F_vector

            def process_interpolation(cond: str,
                                    i: int,
                                    o: int,
                                    interp_data: Union[Tuple[np.ndarray, str],
                                                        Tuple[np.ndarray, np.ndarray, bool, str]]
                                    ) -> None:
                r"""
                Process the interpolation condition for a given component.

                Generates all (j, k) pairs (including the special ('star', 'star') case) and 
                processes them according to the type of interpolation indices.

                :param cond: The condition identifier.
                :param i: The component index.
                :param o: The interpolation index.
                :param interp_data: Interpolation data tuple.
                """
                # Build the complete list of (j, k) pairs, including the star case.
                j_k_pairs_with_star = [(j, k) for j in range(1, algo.m_bar_is[i - 1] + 1) 
                                            for k in range(k_maxs[cond] + 1)]
                j_k_pairs_with_star.append(('star', 'star'))
                
                # Determine the type and extract the appropriate interpolation indices.
                if i in algo.I_op:
                    comp_type = 'op'
                    interpolation_indices = interp_data[1]
                else: 
                    comp_type = 'func'
                    interpolation_indices = interp_data[3]

                if interpolation_indices == 'i<j':
                    for pair1, pair2 in combinations(j_k_pairs_with_star, 2):
                        process_pairs(cond, i, o, interp_data, [pair1, pair2], comp_type)
                elif interpolation_indices == 'i!=j':
                    for pair1, pair2 in product(j_k_pairs_with_star, repeat=2):
                        if pair1 == pair2:
                            continue
                        process_pairs(cond, i, o, interp_data, [pair1, pair2], comp_type)
                elif interpolation_indices == 'i':
                    for pair in j_k_pairs_with_star:
                        process_pairs(cond, i, o, interp_data, [pair], comp_type)
                elif interpolation_indices == 'i!=star':
                    j_k_pairs = [(j, k) for j in range(1, algo.m_bar_is[i - 1] + 1) 
                                        for k in range(k_maxs[cond] + 1)]
                    for pair in j_k_pairs:
                        process_pairs(cond, i, o, interp_data, [pair, ('star', 'star')], comp_type)
                else:
                    raise ValueError(f"Error: Invalid interpolation indices: {interpolation_indices}.")

            # ---------------------------------------------------------------------
            # Loop over all active conditions and components to process interpolation data.
            # ---------------------------------------------------------------------
            for cond in conds:
                for i in range(1, m + 1):
                    # Retrieve the interpolation data for component i.
                    for o, interp_data in enumerate(prob.get_component_data(i)):
                        process_interpolation(cond, i, o, interp_data)

            # ---------------------------------------------------------------------
            # Add final constraints to the model.
            # ---------------------------------------------------------------------
            for cond in conds:
                # Enforce that the PSD constraint sums belong to the PSD cone.
                Mod.constraint(PSD_constraint_sums[cond], Domain.inPSDCone(n + (k_maxs[cond] + 1) * m_bar + m))
                # For functional components, enforce the equality constraint.
                if m_func > 0:
                    Mod.constraint(eq_constraint_sums[cond] == 0)
                
            # ---------------------------------------------------------------------
            # Attempt to solve the model.
            # ---------------------------------------------------------------------
            try:
                Mod.solve()
                Mod.primalObjValue()
            except OptimizeError as e:
                licence_markers = (
                    "err_license_max",         # 1016 – all floating tokens in use
                    "err_license_server",      # 1015 – server unreachable / down
                    "err_missing_license_file" # 1008 – no licence file / server path
                )
                if any(mark in str(e) for mark in licence_markers):
                    raise                    
                return False                 
            except Exception as e:
                # Uncomment the following line for debugging if needed.
                # print("Error during solve: {0}".format(e))
                return False
            return True
    
    @staticmethod
    def _compute_Thetas(algo: Type[Algorithm], h: int, alpha: int, condition: str = 'C1') -> Tuple[np.ndarray, np.ndarray]:
        r"""
        Compute the Theta matrices (capital :math:`\Theta`) using the X matrices.

        For **condition "C1"**:
        
        - Set :math:`k_{\min} = 0` and :math:`k_{\max} = h+\alpha+1`.
        - Retrieve :math:`X = X_{\alpha+1}` from ``algo.get_Xs(k_min, k_max)``.
        - :math:`\Theta_0` is of size :math:`[n + (h+1)\bar{m} + m] \times [n + (h+\alpha+2)\bar{m} + m]`.
        - :math:`\Theta_1` is formed by vertically stacking :math:`X` with a block row 
          consisting of a zero block and an identity matrix.

        For **condition "C4"**:
        
        - Set :math:`k_{\min} = 0` and :math:`k_{\max} = h+\alpha+2`.
        - Retrieve :math:`X = X_1` from ``algo.get_Xs(k_min, k_max)``.
        - :math:`\Theta_0` is of size :math:`[n + (h+\alpha+2)\bar{m} + m] \times [n + (h+\alpha+3)\bar{m} + m]`.
        - :math:`\Theta_1` is formed similarly by stacking :math:`X` with an appropriate block row.

        **Parameters**

        :param algo: An instance of Algorithm (providing n, m_bar, m).
        :param h: A nonnegative integer.
        :param alpha: A nonnegative integer.
        :param condition: Either "C1" or "C4".
        
        **Returns**

        :return: A tuple :math:`(\Theta_0, \Theta_1)`.

        **Raises**

        :raises ValueError: If :math:`h` or :math:`\alpha` is negative or if condition is invalid.
        """
        if h < 0:
            raise ValueError("Parameter h must be >= 0.")
        if alpha < 0:
            raise ValueError("Parameter alpha must be >= 0.")
        if condition not in ('C1', 'C4'):
            raise ValueError("Condition must be either 'C1' or 'C4'.")
        
        n = algo.n
        m_bar = algo.m_bar
        m = algo.m

        if condition == 'C1':
            k_min, k_max = 0, h + alpha + 1
            Xs = algo.get_Xs(k_min, k_max)
            key = alpha + 1
            if key not in Xs:
                raise ValueError(f"Expected key {key} in X matrices, but it was not found.")
            X_mat = Xs[key]
            Theta0 = np.block([
                [np.eye(n + (h + 1) * m_bar), np.zeros((n + (h + 1) * m_bar, (alpha + 1) * m_bar)), np.zeros((n + (h + 1) * m_bar, m))],
                [np.zeros((m, n + (h + 1) * m_bar)), np.zeros((m, (alpha + 1) * m_bar)), np.eye(m)]
            ])
            lower_block = np.hstack([
                np.zeros(((h + 1) * m_bar + m, n + (alpha + 1) * m_bar)),
                np.eye((h + 1) * m_bar + m)
            ])
            Theta1 = np.vstack([X_mat, lower_block])
            return Theta0, Theta1

        elif condition == 'C4':
            k_min, k_max = 0, h + alpha + 2
            Xs = algo.get_Xs(k_min, k_max)
            key = 1
            if key not in Xs:
                raise ValueError(f"Expected key {key} in X matrices, but it was not found.")
            X_mat = Xs[key]
            Theta0 = np.block([
                [np.eye(n + (h + alpha + 2) * m_bar), np.zeros((n + (h + alpha + 2) * m_bar, m_bar)), np.zeros((n + (h + alpha + 2) * m_bar, m))],
                [np.zeros((m, n + (h + alpha + 2) * m_bar)), np.zeros((m, m_bar)), np.eye(m)]
            ])
            lower_block = np.hstack([
                np.zeros(((h + alpha + 2) * m_bar + m, n + m_bar)),
                np.eye((h + alpha + 2) * m_bar + m)
            ])
            Theta1 = np.vstack([X_mat, lower_block])
            return Theta0, Theta1

        # Should never reach here.
        raise ValueError("Unexpected error in _compute_Thetas.")

    @staticmethod
    def _compute_thetas(algo: Type[Algorithm], h: int, alpha: int, condition: str = 'C1') -> Tuple[np.ndarray, np.ndarray]:
        r"""
        Compute the theta matrices (lowercase :math:`\theta`) for functional evaluations.

        For **condition "C1"**:
        
        - :math:`\theta_0 \in \mathbb{R}^{((h+1)\bar{m}_{\text{func}}+m\_func) \times ((h+\alpha+2)\bar{m}_{\text{func}}+m\_func)}` 
          is given by a block matrix with an identity in the upper left and lower right.
        - :math:`\theta_1` is formed as a horizontal block consisting of a zero block and an identity matrix.

        For **condition "C4"**:
        
        - :math:`\theta_0 \in \mathbb{R}^{((h+\alpha+2)\bar{m}_{\text{func}}+m\_func) \times ((h+\alpha+3)\bar{m}_{\text{func}}+m\_func)}` 
          is defined similarly.
        - :math:`\theta_1` is a horizontal block with a zero block and an identity matrix.

        **Parameters**

        :param algo: An instance of Algorithm (providing :math:`\bar{m}_{\text{func}}` and :math:`m\_func`).
        :param h: A nonnegative integer.
        :param alpha: A nonnegative integer.
        :param condition: Either "C1" or "C4".

        **Returns**

        :return: A tuple :math:`(\theta_0, \theta_1)`.

        **Raises**

        :raises ValueError: If :math:`h` or :math:`\alpha` is negative, if condition is invalid,
                            or if there are no functional components (i.e. :math:`m\_func \le 0`).
        """
        if h < 0:
            raise ValueError("Parameter h must be >= 0.")
        if alpha < 0:
            raise ValueError("Parameter alpha must be >= 0.")
        if condition not in ('C1', 'C4'):
            raise ValueError("Condition must be either 'C1' or 'C4'.")
        
        m_bar_func = algo.m_bar_func
        m_func = algo.m_func
        
        # Theta matrices are only defined when there is at least one functional component.
        if m_func <= 0:
            raise ValueError("Theta matrices require at least one functional component (m_func > 0).")
        
        if condition == 'C1':
            theta0 = np.block([
                [np.eye((h + 1) * m_bar_func), np.zeros(((h + 1) * m_bar_func, (alpha + 1) * m_bar_func)), 
                np.zeros(((h + 1) * m_bar_func, m_func))],
                [np.zeros((m_func, (h + 1) * m_bar_func)), np.zeros((m_func, (alpha + 1) * m_bar_func)), 
                np.eye(m_func)]
            ])
            theta1 = np.hstack([
                np.zeros(((h + 1) * m_bar_func + m_func, (alpha + 1) * m_bar_func)),
                np.eye((h + 1) * m_bar_func + m_func)
            ])
            return theta0, theta1

        elif condition == 'C4':
            theta0 = np.block([
                [np.eye((h + alpha + 2) * m_bar_func), np.zeros(((h + alpha + 2) * m_bar_func, m_bar_func)), 
                np.zeros(((h + alpha + 2) * m_bar_func, m_func))],
                [np.zeros((m_func, (h + alpha + 2) * m_bar_func)), np.zeros((m_func, m_bar_func)), 
                np.eye(m_func)]
            ])
            theta1 = np.hstack([
                np.zeros(((h + alpha + 2) * m_bar_func + m_func, m_bar_func)),
                np.eye((h + alpha + 2) * m_bar_func + m_func)
            ])
            return theta0, theta1

        # Should never reach here.
        raise ValueError("Unexpected error in _compute_thetas.")


from pathlib import Path
from setuptools import setup, find_packages

here = Path(__file__).parent
long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name="autolyap",
    version="0.1.0",
    author="Manu Upadhyaya",
    author_email="manu.upadhyaya.42@gmail.com",
    description="Automatic Lyapunov analysis",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://autolyap.github.io",
    project_urls={
        "Documentation": "https://autolyap.github.io/",
        # later, add "Source": "https://github.com/…"
    },
    license="MIT",
    python_requires=">=3.9",
    packages=find_packages(include=["autolyap", "autolyap.*"]),
    install_requires=[
        "numpy>=1.24",
        "mosek>=10.0"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering :: Mathematics"
    ],
)



import numpy as np
from typing import Tuple
from .algorithm import Algorithm

class DavisYin(Algorithm):
    # https://arxiv.org/pdf/1504.01032.pdf
    def __init__(self, gamma, lambda_value):
        super().__init__(1, 3, [1, 1, 1], [1, 2, 3], [])
        self.gamma = gamma
        self.lambda_value = lambda_value
    
    def set_gamma(self, gamma):
        self.gamma = gamma

    def set_lambda(self, lambda_value):
        self.lambda_value = lambda_value
    
    def get_ABCD(self, k: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        A = np.array([[1]])
        B = -self.gamma*self.lambda_value*np.array([[1,1,1]])
        C = np.array([[1],[1],[1]])
        D = -self.gamma*np.array([[1,0,0],[1,0,0],[2,1,1]])
        return (A, B, C, D)